{: XMLReader
PURPOSE AND IMPLEMENTATION:
	This unit defines classes for parsing a well-formed standalone XML document in
	to a tree of nodes describing the data therein.  These classes only support a
	subset of the XML 1.0 recommendation and have not been analyzed for
	conformance with the recommendation.  This module also only supports 8-bit
	character sets and makes no special allowances for UTF-8.  Nor has any support
	for the "<!DOCTYPE" node been implemented.  Parsing of the XML declaration
	lazily uses the attribute parsing code which is not correct (but works).  This
	module also freely trims element whitespace which is not always correct.

	Q:  Why use this module instead of Microsoft's XML parser or the OpenXML
			project?

	A:  I wrote this module to be free of dependencies on outside libraries and to
			be small and fast.  Neither of the above libraries met that criteria.  I
			didn't like the idea of requiring my users to install Microsoft's XML
			library (and the potential compatibility problems that arise).  Nor did I
			like the idea of adding 135 kb (OpenXML) to my executable just to load files
			in XML format.  Instead this unit only uses about 7 kb.

	This was written for Delphi 5.  It may work for other versions, but does
	require method overloading which was introduced in Delphi 4.

HISTORY:
	09/12/00 - Created by Colin Patrick Sarsfield

DISCLAIMER:
	Use at your own risk.

COPYRIGHT:
	Released into the public domain by Colin Sarsfield 9/13/00
}

{: XMLWriter
PURPOSE AND IMPLEMENTATION:
	This unit defines classes for writing data out to a well-formed standalone XML
	document.  These classes only support a subset of the XML 1.0 recommendation
	and have not been analyzed for conformance with the recommendation.  This
	module also only supports 8-bit character sets and makes no special allowances
	for UTF-8.

	This was written for Delphi 5.  It may work for other versions, but does
	require method overloading which was introduced in Delphi 4.

HISTORY:
	09/11/00 - Created by Colin Patrick Sarsfield

DISCLAIMER:
	Use at your own risk.

COPYRIGHT:
	Released into the public domain by Colin Sarsfield 9/13/00
}

{: XMLNode

	Original XMLReader and XMLWriter from Colin repackaged by Tomislav Karda (tomislav.kardas@pmx.hr)
	into one XMLNode unit and a lot of changes/improvements made to original code.

	- TMXLNode class improved
			- parent, data, text, exacttext, xml properties added
			- Attributes and Elements	array properties added
			- TXMLNodeType extended for two new node types: xntText, xntCData to support Mixed Content
			- some things removed, like saving node data AsFloat, AsDateTime,
				etc. because it depends on system settings. Only string access left.
			- tried to follow names that MSXML uses for properties and method of the node object

	- improved XML readers
			- from string, stream and file reader
			- events added to XMLReader so that parsing can be intercepted when node completes, etc.
			- doctype parsing added just to store the info in the node, not to abort parsing of such documents
			- PreserveWhiteSpace, MixedContent, ProgressPercent, Size, Position - properties added to reader
			- Text and CData node types added, and Mixed Content parsing supported. xntText & xntCData nodes
				implemented in the ParseContent	procedure where for each text or CData section new node	is created
				instead of accumulating	all text into element node.

	- improved XML writers
			- to string, stream and file writer
			- events added to XMLWriter so that writing can be intercepted on certain events.
			- FormatOutput, ProgressPercent, Position, Total, Count - properties added to writer
			- WriteNode procedure Added to XMLWriter for full node write in one method call

	- XML reader & writer fully optimized for Speed and Memory requirement

	- different compile options switches added

	- TXMLDocument component added for easy using of XMLNode, XMLWriter and XMLReader objects through
		simple methods for loading and saving from/to different sources/destination and using
		events to accomplish progress counter or alter the process of load/save.

	- Demo application improved	
}

unit XMLNode;

interface

uses
	Classes, SysUtils;

(*
	Compile options

	OBJBUF - Specifies how are Name & Date properties of TXMLNode object handled inside
		0	- as string variables, little faster execution, little bigger memory requirements
		1	- as PChar buffers, little slower execution, little less memory requirements
		Note: Either option works OK. But I guess turning off this option is more natural way, it looks
					little weird to have members handled that way when just plain string variables will do. And
					also, I would give speed a little preference over memory.

	LIST - Specifies how is list of child nodes implemented
		0	- Using TList object, automaticaly created when Node object is created
		1	- Using TList object, created only when Node object has child nodes
		2	- Without TList object, using double linked lists, this increase
				memory usage a little, but gains a lot on speed because new objects are inserted at the end of the list
				by just changing a few pointers. There is no need to maintain separate TList object to hold pointers
				to every child node. This also gains in speed since there is no need to realocate TList memory when number
				of child nodes increases. But this also imposes a different programming style since there is no index
				approach to child nodes, rather you can approach to first and last child nodes and then skip forward
				and backward. I prefere this way since it's faster and more natural.

	PARENT - Specifies if node object has a reference to it's parent or not
		0	- there is no Parent property of Node object
		1	- there is a Parent property of Node object
		Note: This option does not affect performance. OK, it may be faster and less memory required if
					turned off because there is no need to deal with additional property, but having parent node
					property seams like a very handy thing to me. Especialy when using OnAfterNodeRead event when you
					get the TXMLNode object just completed and you have to process it and the parent node information
					might be relevant.

	PARSEBUF - Specifies if buffer is used whileparsing
		0	- The worst option, buffer is not used, string variables serve as temporary buffers
		1 - Reading of node elements is buffered through PChar buffers. This option is added because
				otherwise strings take a lot of virtual memory and specialy when large files are parsed (few MB),
				it also increases the parsing speed rapidly. Virtual memory requirements dramaticaly reduced
				since there is no string handling around.
		Note: You should never ever have a need to turn off this option, except for educational purposes to see
					how much speed is lost in that way.

	UNREADBUF - Specifies if unread buffer is used whileparsing to store chars that have been read but returned back
		0	- No unread buffer, parser uses SaveState and RestoreState technique
		1	-	Better policy. Chars are pulled out from the xml source and if parser needs to go back it returns
				chars back (unreads them) and they are stored into the unread buffer. Subsequent reads first check
				the unread buffer and then if empty pulls the xml source.

	Options for best performance:	OBJBUF0, LIST2, PARSEBUF1

	Options PARENT1 and UNREADBUF1 aren't realy for performance but alter the implementation logic, I recomend them.
*)

{$define OBJBUF0}
{$define LIST2}
{$define PARENT1}
{$define PARSEBUF1}
{$define UNREADBUF1}

type
  EXmlError = class(Exception);

	TXBuf = object
	private
		FSize: integer;
		FPtr: PChar;
		function GetAsString: string;
		procedure SetAsString(value: string);
		procedure SetSize(value: integer);
	public
		procedure Init;
		procedure Clear;
		procedure CopyFrom(SourcePtr: PChar; SourceSize: integer);
		procedure AppendFrom(SourcePtr: PChar; SourceSize: integer);
		property AsString: string read GetAsString write SetAsString;
		property Size: integer read FSize write SetSize;
		property Ptr: PChar read FPtr;
	end;

type
	CharSet = set of Char;

	TXMLNodeType = (
		xntDocument,
		xntXMLDeclaration,
		xntDocumentType,
		xntElement,
		xntAttribute,
		xntText,
		xntCData,
		xntProcessingInstruction
	);

	(*
		There is no programmed limitation to what node types can be childs to other node types,
		but users should hold to following logic which is supported by Parser and Writter:

		xntDocument:
			xntXMLDeclaration
			xntDocumentType
			xntProcessingInstruction
		xntXMLDeclaration:
			xntAttribute
		xntElement:
			xntAttribute
			xntText
			xntCData
			xntElement
			xntProcessingInstruction
	*)

const
	TXMLNodeTypeStrings: array[TXMLNodeType] of String = (
		'Document',
		'XMLDeclaration',
		'DocumentType',
		'Element',
		'Attribute',
		'Text',
		'CDataSection',
		'ProcessingInstruction'
	);

type
	TXMLNode = class
	private
{$ifdef PARENT1}
		FParentNode: TXMLNode;
{$endif PARENT1}
		FNodeType: TXMLNodeType;

{$ifdef OBJBUF1}
		FNodeName: TXBuf;
		FNodeValue: TXBuf;
{$else}
		FNodeName: String;
		FNodeValue: String;
{$endif OBJBUF1}

{$ifdef LIST2}
		FPreviousSibling: TXMLNode;			// Previous sibling
		FNextSibling: TXMLNode;			// Next sibling
		FChildNodeCount: integer;	// Number of child nodes
		FFirstChild: TXMLNode;	// First child node
		FLastChild: TXMLNode;	// Last child node
{$else}
		FNodes: TList;
{$endif LIST2}

{$ifdef PARENT1}
		procedure SetParent(AParent: TXMLNode);
		function GetOwnerDocument: TXMLNode;
		function GetPath: string;
{$endif PARENT1}
{$ifdef OBJBUF1}
		function GetNodeName: string;
		procedure SetNodeName(value: string);
		function GetNodeValue: string;
		procedure SetNodeValue(value: string);
{$endif OBJBUF1}
{$ifndef LIST2}
		function GetChildNodeCount: Integer;
		function GetNodes(Index: Integer): TXMLNode;
{$endif LIST2}
		function GetTotalNodeCount: Integer;

		function GetAttributes(Name: string): TXMLNode;
		function GetElements(Name: string): TXMLNode;

		function GetData: string;
		procedure SetData(value: string);
		function GetText: string;
		function GetExactText: string;
		function GetXml: string;
	public
		// default NodeType is xntElement, Name and Data are ''
		constructor Create(const NodeType: TXMLNodeType = xntElement);
		// when destroyed, node will destroy all it's child nodes
		destructor Destroy; override;

{$ifdef OBJBUF1}
		property NodeName: String read GetNodeName write SetNodeName;
		property NodeValue: String read GetNodeValue write SetNodeValue;
{$else}
		property NodeName: String read FNodeName write FNodeName;
		property NodeValue: String read FNodeValue write FNodeValue;
{$endif OBJBUF1}
		property NodeType: TXMLNodeType read FNodeType write FNodeType;
		function Description: String;

{$ifdef PARENT1}
		// ParentNode:
		//      	Represents node that is a parent to the node. ParentNode property is automaticaly
		//				set by using NewChild & AppendChild methods, and cleared when using RemoveChild
		//				on the parent node.
		//				If assigned directly it will remove the node from previous parent and add it as a
		//				child to new parent if new parent is not nil.
		//	OwnerDocument:
		//				Document that contains this node. It's a root node that has no parent.
		//	Path:
		//				Path to this node made of NodeNames of parent nodes starting from the root node 
		property ParentNode: TXMLNode read FParentNode write SetParent;
		property OwnerDocument: TXMLNode read GetOwnerDocument;
		property Path: String read GetPath;
{$endif PARENT1}

{$ifdef LIST2}
		// FirstChild:
		//      	First child node object to this node.
		// LastChild:
		//      	Last child node object to this node.
		// PreviousSibling:
		//      	Previous sibling node object to this node. This is a node that has same parent as this node but
		//				is right before him in the list of child nodes for it's parent.
		// NextSibling:
		//      	Next sibling node object to this node. This is a node that has same parent as this node but
		//				is right after him in the list of child nodes for it's parent.
		// ChildNodeCount:
		//      	Number of child nodes for this node.
		property FirstChild: TXMLNode read FFirstChild;
		property LastChild: TXMLNode read FLastChild;
		property PreviousSibling: TXMLNode read FPreviousSibling;
		property NextSibling: TXMLNode read FNextSibling;
		property ChildNodeCount: Integer read FChildNodeCount;
{$else}
		property ChildNodeCount: Integer read GetChildNodeCount;
		property Nodes[Index: Integer]: TXMLNode read GetNodes;
{$endif LIST2}

		// TotalNodeCount:
		// 				Total Node Count, including all child nodes in the three below this node
		property TotalNodeCount: Integer read GetTotalNodeCount;

		// CreateChild:
		//				Creates new child node object
		function CreateChild(const NodeType: TXMLNodeType = xntElement): TXMLNode;
{$ifdef LIST2}
		// AppendChild:
		//				Adds existing node object as a child for current node,
		// 				if node is a child of other parent, reference is automaticaly removed
		// RemoveChild:
		//				removes node from the list of child nodes,
		//				node's parent is set to nil, and node is not destroyed.
		//				returns true if node was a child node.
		//				Note: To destroy a child node simple call it's free method,
		//				this will remove it from the child list and destroy
		procedure AppendChild(Node: TXMLNode);
		function RemoveChild(Node: TXMLNode): boolean;
{$else}
		// AppendChild:
		//				Adds existing node object as a child for current node, returns node index
		// 				if node is a child of other parent, reference is automaticaly removed
		// RemoveChild:
		//				removes node from the list of child nodes,
		//				node's parent is set to nil, and node is not destroyed.
		//				returns node's index from a child list, -1 if node not found
		//				in the list and couldn't been removed.
		//				Note: To destroy a child node simple call it's free method,
		//				this will remove it from the child list and destroy
		// Delete:
		//				removes node at index from the list of child nodes,
		//				node's parent is set to nil, and node is not destroyed.
		//				returns node object reference that application has to use or destroy,
		//				otherwise memory loss will occure.
		//				Note: To destroy a child node simple call it's free method,
		//				this will remove it from the child list and destroy
		// IndexOf:
		//				returns node index, -1 if not found
		function AppendChild(Node: TXMLNode): Integer;
		function RemoveChild(Node: TXMLNode): integer;
		function Delete(Index: Integer): TXMLNode;
		function IndexOf(Node: TXMLNode): integer;
{$endif LIST2}

		// CloneNode:
		//			 		Clones node and makes a new copy of it. The clone has no parent node and it is a duty of
		//					application to append it as client to some node or to free it manualy when not needed.
		//					If deep is set then all the subtree is cloned together with this node.
		function CloneNode(deep: boolean): TXMLNode;

		function HasChildNodes: boolean;

		// FindChild:
		//					finds n'th (Index) occurence of a node with given Name and NodeType.
		//					if node not found returns nil, or if CreateMissing specified then
		//					new node is created and returned as found.
		function FindChild(const NodeName: String; const NodeType: TXMLNodeType = xntElement;
			const Index: Integer = 0; const CreateMissing: boolean = false): TXMLNode;

		// FindChildExc:
		//					finds n'th (Index) occurence of a node with given Name and NodeType.
		//					if node not found, an exception is raised.
		function FindChildExc(const NodeName: String; const NodeType: TXMLNodeType = xntElement;
			const Index: Integer = 0): TXMLNode;

		// Attributes:	access to node of type xntAttribute via node name
		// Elements:		access to node of type xntElement via node name
		property Attributes[Name: string]: TXMLNode read GetAttributes;
		property Elements[Name: string]: TXMLNode read GetElements;

		// Data:
		//					When read, Data property gives NodeValue plus all NodeValues of child Text and CData nodes.
		//					When write, Data property stored NodeValue and destroys all child Text and CData nodes.
		//					Data property gives a flexible and easier approach to simple elements and eliminates the trouble
		//					to read element node values when document is loaded with MixedContent = true. Using this Data
		//					property application doesn't have to worry if element value is actualy stored in the element node
		//					itself or in it's child Text and CData nodes. It's whole purpose is for DataSet fields mapping to
		//					xml element nodes to simplify processing of reading and writing field values from/to element nodes.
		// Text:
		//					Gives text representation of the node. Text is built from this node and all the child Element,
		//					Text and CData nodes by adding theirs NodeValues. The resulting text is trimmed of white space
		//					from the beggining and end. Special handling is added to detect situations where two consequtive
		// 					non empty element nodes are combined in one text string to separate them with space.
		// 					This might not be the best approach, since when white space is significant (like in html)
		//					this adds extra space where it should not.
		//					The example of unwanted behaviour is: '<NODE><U>First</U><I><B>Second</B></I></NODE>'
		//					Where Text would be: 'First Second' while actually 'FirstSecond' should be returned.
		//					For such application use ExactText property.
		// ExactText:
		//					The same as Text, but no extra space added between two consequtive non empty element nodes.
		// Xml:
		//					gives the xml string of the subtree

		property Data: String read GetData write SetData;
		property Text: String read GetText;
		property ExactText: String read GetExactText;
		property Xml: String read GetXml;

		(*
			It could also be implemented, but first TXMLNodeList has to be realized as a descendant from TList
			and then some kind of a search engine upon the node tree structure can be built with following
			function prototypes.

		function getElementsByTagName(const tagName: String): TXMLNodeList;
		function selectNodes(const queryString: String): TXMLNodeList;
		function selectSingleNode(const queryString: String): TXMLNode;
		*)
	end;

{$ifdef UNREADBUF0}
type
	TXMLReaderState = record
		Position: Integer;
		CurChar: Char;
		// If additional state data will be required in some Reader descendant implementations,
		// it can be implemented here as a kind of a generic Pointer and can be allocated in
		// Reader descendant class.
		// For String & Stream readers only Position and Char is enough, so it is better
		// to realize TXMLReaderState as record then a TObject descendant since it would generate
		// huge alloc/free memory calls for every Save and Free state in TObject implementation.
		// FreeState procedure is removed, for speed
	end;
{$endif UNREADBUF0}

const
(*
	XMLStreamReaderCacheSize -  Cache size (64 KB)
	XMLStreamReaderCacheSave -  Min. size of behind cache (256 B)
	XMLStreamReaderLoadMask  -  Mask for rounding data size loading from stream, round to 1KB
*)
	XMLStreamReaderCacheSize  = $010000;
{$ifdef UNREADBUF0}
	XMLStreamReaderCacheSave  = $000100;
	XMLStreamReaderLoadMask   = $FFFC00;
{$endif UNREADBUF0}

type
	TXMLReader = class
	private
		CurChar: Char;

		FPreserveWhiteSpace: boolean;
		FMixedContent: boolean;
		FProgressStep: integer;

		(*
			FProgressCount:
				Number of bytes that has to be read before calling each OnProgress event
			FProgressPos:
				Current countdown of read bytes before calling next OnProgress event
		*)
		FProgressCount: integer;
		FProgressPos: integer;
		FProgressPercent: integer;
		FNode: TXMLNode;

		FOnNode: TNotifyEvent;
		FOnElementStart: TNotifyEvent;
		FOnElementEnd: TNotifyEvent;
		FOnProgress: TNotifyEvent;

{$ifdef PARSEBUF1}
	private
		(*
			Read buffer intruduced here to avoid using string type in character data processing
			since string + char operation takes much more time than storing byte in a buffer

			ReadBufAdd:
							adds one char to read buff.
			ReadBufClose:
							closes read buffer from the given position and returns buffered string that has accumulated
							from that position.
		*)
		FReadBuf: TXBuf;
		FReadBufPos: integer;
		procedure ReadBufAdd(c: char);
		function ReadBufClose(p: integer): TXBuf;
{$endif PARSEBUF1}

{$ifdef UNREADBUF1}
	private
		(*
			UnRead buffer intruduced here to avoid requesting xml source to go backwards when needed. Temporary
			chars that have been read will be placed here for next char reading.

			UnReadChar:
							Unreads one character.
			UnReadString:
							Unreads a character sequence. Note here that character sequences are unread in reverse order
							(from last to first char) so that when reading again it will read the same string. Unreading
							is a LIFO (last in first out) structure.
		*)
		FUnReadBuf: TXBuf;
		FUnReadBufPos: integer;
		procedure UnReadChar(c: char);
		procedure UnReadString(p: PChar; count: integer);
{$endif UNREADBUF1}

		function DoesStreamMatchString(Str: String): Boolean;
		procedure ReadCharAndCheckForEnd;
		function ParseAttribute(Node: TXMLNode): Boolean;
		function ParseAttValue(var AttValue: String): Boolean;
		procedure ParseComment;
		procedure ParseContent(Node: TXMLNode; EndTagName: string);
		procedure ParseElement(Node: TXMLNode);
		function ParseEq: Boolean;
{$ifdef PARSEBUF1}
		function ParseName(var Name: TXBuf): boolean;
{$else}
		function ParseName(var Name: string): boolean;
{$endif PARSEBUF1}
		procedure ParseProcessingInstruction(Node: TXMLNode);
		function ParseReference: Char;
		function ParseWhitespace: Boolean;
		procedure ReportError(Err: String);
	protected
		procedure DoNode(Node: TXMLNode);
		procedure DoElementStart(Node: TXMLNode);
		procedure DoElementEnd(Node: TXMLNode);
		procedure DoProgress;

		function GetSize: integer; virtual; abstract;
		function GetPosition: integer; virtual; abstract;
		function EndOfData: Boolean; virtual; abstract;
{$ifdef UNREADBUF1}
		procedure ReadChar;
		function GetChar: char; virtual; abstract;
{$else}
		procedure ReadChar; virtual;
		procedure SaveState(var State: TXMLReaderState); virtual; abstract;
		procedure RestoreState(var State: TXMLReaderState); virtual; abstract;
{$endif UNREADBUF1}
	public
		constructor Create;
		destructor Destroy; override;

		// PreserveWhiteSpace:
		// 				If set then all the white space that appears between elements will be parsed
		//				into the document. Documents that were written with FormatOutput = true, should be read with
		// 				PreserveWhiteSpace set to false so that whitespace used for output formating would not be parsed
		//				in and waist memory and create unnecesary text objects.
		//				This property can be changed in the reader events and in that way alter the parsing of different nodes.
		// MixedContent:
		// 				If not set then all the text and CData inside the node goes directly to NodeValue of the element.
		//				Otherwise text and CData are stored in separate child nodes of the element node. If XML document
		//				that is beeing read is known to not have mixed content then this property can be turned off
		//				to not create separate node for every text and CData. There is never a problem if you always leave
		//				MixedContent = true and use Data property of the element node to read simple elements.
		//				This property can be changed in the reader events and in that way alter the parsing of different nodes.
		// ProgressStep:
		//				Percent step value, how many percents of job done have to change to call OnProgress event,
		//				default is 5 what means that OnProgress event is called on 0, 5, 10, 15, ... 100 percent of progress.
		//				If this property need to be changed, it only takes effect if set before calling of ParseDocument, do
		//				not change it in reader events, since that would make ProgressPercent property out of sense.
		property PreserveWhiteSpace: boolean read FPreserveWhiteSpace write FPreserveWhiteSpace default false;
		property MixedContent: boolean read FMixedContent write FMixedContent default false;
		property ProgressStep: integer read FProgressStep write FProgressStep default 5;

		// ParseDocument:
		// 				Fills the XML document structure from the XMLReader source
		procedure ParseDocument(Document: TXMLNode);

		// Size:
		// 				Total size in bytes of XML text in XMLReader source
		// Position:
		// 				Current reader position in XML text in XMLReader source
		// ProgressPercent:
		// 				Current percent of read job done, from 0-100 in step of 5
		// Node:
		// 				Current Node when OnNode, OnElementStart, OnElementEnd executes
		property Size: integer read GetSize;
		property Position: integer read GetPosition;
		property ProgressPercent: integer read FProgressPercent;
		property Node: TXMLNode read FNode;

		property OnNode: TNotifyEvent read FOnNode write FOnNode;
		property OnElementStart: TNotifyEvent read FOnElementStart write FOnElementStart;
		property OnElementEnd: TNotifyEvent read FOnElementEnd write FOnElementEnd;
		property OnProgress: TNotifyEvent read FOnProgress write FOnProgress;
	end;

	TXMLStringReader = class(TXMLReader)
	private
		FData: String;
		FPosition: Integer;
	protected
		function GetSize: integer; override;
		function GetPosition: integer; override;
		function EndOfData: Boolean; override;
{$ifdef UNREADBUF1}
		function GetChar: char; override;
{$else}
		procedure ReadChar; override;
		procedure SaveState(var State: TXMLReaderState); override;
		procedure RestoreState(var State: TXMLReaderState); override;
{$endif UNREADBUF1}
	public
		constructor Create(AData: string);
		property Data: String read FData;
	end;

	TXMLStreamReader = class(TXMLReader)
	private
		FStream: TStream;
		FSize: integer;			// Stream size
		FPosition: integer;	// Current stream position

		FCacheBuf: array [0..XMLStreamReaderCacheSize-1] of char; // Stream cache
		FCacheLen: integer; // How many chars is in the cache

		// How many chars we are behind the current stream position inside the cache,
		// if 0 then we have exhausted the cache and have to read from stream, if not read from cache.
		FBehind: integer;
	protected
		function GetSize: integer; override;
		function GetPosition: integer; override;
		function EndOfData: Boolean; override;
{$ifdef UNREADBUF1}
		function GetChar: char; override;
{$else}
		procedure ReadChar; override;
		procedure SaveState(var State: TXMLReaderState); override;
		procedure RestoreState(var State: TXMLReaderState); override;
{$endif UNREADBUF1}
	public
		constructor Create(AStream: TStream);
		property Stream: TStream read FStream;
	end;

	TXMLFileReader = class(TXMLStreamReader)
	private
		FFileName: string;
	protected
	public
		constructor Create(AFileName: string);
		destructor Destroy; override;
		property FileName: string read FFileName;
	end;

const
	XMLWriterCacheSize = $010000;

type
	TXMLTagType = (xttStarting, xttStartingAndEnding, xttEnding, xttEndingSameLine);

	TXMLWriter = class
	private
		FWriteBuf: TXBuf;
		FWriteBufPos: integer;
		FIndentation: Integer;
		FFirstLine: boolean;

		FFormatOutput: boolean;
		//FProgressStep: integer;

		FPosition: integer;
		FTotal: integer;
		FCount: integer;
		FProgressPercent: integer;
		FNode: TXMLNode;

		FOnElementStart: TNotifyEvent;
		FOnElementEnd: TNotifyEvent;
		FOnProgress: TNotifyEvent;
	protected
		procedure WriteBuffer(Buffer: PChar; Count: integer); virtual;
		procedure WriteChar(c: char);

		procedure DoElementStart(Node: TXMLNode);
		procedure DoElementEnd(Node: TXMLNode);
		procedure DoProgress;
	public
		constructor Create;
		destructor Destroy; override;

		// FormatOutput
		//				If set then white space is added so that document output looks nice when viewed/edited
		//				as text file. This can be done only if white space is not important in document content
		//				and readings should be parsed in XMLReader with PreserveWhiteSpace = false.
		property FormatOutput: boolean read FFormatOutput write FFormatOutput default true;

		// ProgressStep:
		//				Percent step value, how many percents of job done have to change to call OnProgress event,
		//				default is 5 what means that OnProgress event is called on 0, 5, 10, 15, ... 100 percent of progress.
		//				If this property need to be changed, it only takes effect if set before calling of WriteNode, do
		//				not change it in writer events, since that would make ProgressPercent property out of sense.
		//property ProgressStep: integer read FProgressStep write FProgressStep default 5;

		// Position
		//				Number of characters already written to the Writer destination. There is no Size property
		//				since Size is not known until end, but Total & Count are used for that purpose since
		//				that gives an idea of where we are right now in the process of writing.
		// Total
		//				Total number of nodes that has to be written to the destination. It is set in WriteNode
		//				method when invoked after StartDoc.
		// Count
		//				Number of nodes that are already written to the destination.
		// ProgressPercent
		//				Percent of write job done. It is calculated as: (100 * Count) div Total.
		// Node:
		// 				Current Node when OnElementStart, OnElementEnd executes
		property Position: integer read FPosition;
		property Total: integer read FTotal;
		property Count: integer read FCount;
		property ProgressPercent: integer read FProgressPercent;
		property Node: TXMLNode read FNode;

		// StartDoc
		//				Prepares writter for writing.
		// WriteNode
		//				Writes complete node with all child elements to destination.
		// EndDoc
		//				Finishes writting.
		//
		// Note:	The easiest way for application to write XML document is to construct TXMLNode object that
		//				represents whole Document, fill it with appropriate child elements and then call:
		//				StartDoc, WriteNode, EndDoc. Other write methods below are automaticaly called by WriteNode
		//				but can be used directly if application wants to write without constructed TXMLNode document
		//				object or if only parts of XML document are constructed in TXMLNode objects. WriteNode can be
		//				mixed with other Write methods.
		procedure StartDoc;
		procedure WriteNode(Node: TXMLNode);
		procedure EndDoc;

		procedure WriteStr(const Value: String);
		procedure Write(const Value: String);
		procedure WriteCDATA(Data: string);
		procedure WriteData(const Name: String; const Data: String);
		procedure WriteStandaloneDocumentDeclaration;
		procedure WriteTag(const Name: String; TagType: TXMLTagType = xttStarting; Open: Boolean = False);
		procedure WriteTagParam(const ParamName, Value: String);
		procedure WriteTagClose(TagType: TXMLTagType = xttStarting);
		// WriteLineEnd does nothing if FormatOutput = false, otherwise outputs newline (CRLF) and idents
		// with spaces according to the open tag nesting level.
		procedure WriteLineEnd;

		property OnElementStart: TNotifyEvent read FOnElementStart write FOnElementStart;
		property OnElementEnd: TNotifyEvent read FOnElementEnd write FOnElementEnd;
		property OnProgress: TNotifyEvent read FOnProgress write FOnProgress;
	end;

	TXMLStringWriter = class(TXMLWriter)
	private
		FString: PString;
	protected
		procedure WriteBuffer(Buffer: PChar; Count: integer); override;
	public
		constructor Create(var AString: string);
	end;

	TXMLStreamWriter = class(TXMLWriter)
	private
		FStream: TStream;
	protected
		procedure WriteBuffer(Buffer: PChar; Count: integer); override;
	public
		constructor Create(AStream: TStream);
		property Stream: TStream read FStream;
	end;

	TXMLFileWriter = class(TXMLStreamWriter)
	private
		FFileName: string;
	public
		constructor Create(AFileName: string);
		destructor Destroy; override;
		property FileName: string read FFileName;
	end;

type
	TXMLReaderEvent = procedure (Sender: TObject; Reader: TXMLReader) of object;
	TXMLWriterEvent = procedure (Sender: TObject; Writer: TXMLWriter) of object;

type
	TXMLDocument = class(TComponent)
	private
		FDocument: TXMLNode;
		FIsEmpty: boolean;
		FLoading: boolean;
		FSaving: boolean;

		FPreserveWhiteSpace: boolean;
		FMixedContent: boolean;
		FProgressStep: integer;
		FFormatOutput: boolean;

		FOnReadDocumentStart: TXMLReaderEvent;
		FOnReadDocumentEnd: TXMLReaderEvent;
		FOnReadNode: TXMLReaderEvent;
		FOnReadElementStart: TXMLReaderEvent;
		FOnReadElementEnd: TXMLReaderEvent;
		FOnReadProgress: TXMLReaderEvent;

		FOnWriteDocumentStart: TXMLWriterEvent;
		FOnWriteDocumentEnd: TXMLWriterEvent;
		FOnWriteElementStart: TXMLWriterEvent;
		FOnWriteElementEnd: TXMLWriterEvent;
		FOnWriteProgress: TXMLWriterEvent;

		procedure	DoReadDocumentStart(Reader: TXMLReader);
		procedure	DoReadDocumentEnd(Reader: TXMLReader);
		procedure	DoReadNode(Sender: TObject);
		procedure	DoReadElementStart(Sender: TObject);
		procedure	DoReadElementEnd(Sender: TObject);
		procedure	DoReadProgress(Sender: TObject);

		procedure	DoWriteDocumentStart(Writer: TXMLWriter);
		procedure	DoWriteDocumentEnd(Writer: TXMLWriter);
		procedure	DoWriteElementStart(Sender: TObject);
		procedure	DoWriteElementEnd(Sender: TObject);
		procedure	DoWriteProgress(Sender: TObject);

		procedure CheckBusy;
		procedure LoadFromReader(r: TXMLReader);
		procedure SaveToWriter(w: TXMLWriter);
	public
		constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;

		// Document:
		// 				Document node that contains all the nodes, including XML declaration, root document element, etc...
		//				this is a complete XML file in tree object model that can either be filled out programaticaly
		//				using	CreateNode method or loaded from some XML source.
		property Document: TXMLNode read FDocument;

		// Clear:
		// 				Frees any nodes from Document node. Complete clear, as was after creating a component.
		procedure Clear;

		(*
		property DocumentElement: TXMLNode read GetDocumentElement write SetDocumentElement;
		function createNode(NodeType: TXMLNodeType; const Name: String): TXMLNode;
		function createElement(const Name: String): TXMLNode;
		function createAttribute(const Name: String): TXMLNode;
		*)

		// LoadFromXXX:
		//        Loads a complete XML Document into Document object model into Document property,
		//				any previous Document object tree is removed and new object model from source is placed.
		procedure LoadFromString(xmlstring: string);
		procedure LoadFromStream(xmlstream: TStream);
		procedure LoadFromFile(xmlfilename: string);

		// SaveToXXX:
		//        Saves complete XML Document in Document object model from Document property into
		//				destination media.
		function SaveToString: string;
		procedure SaveToStream(xmlstream: TStream);
		procedure SaveToFile(xmlfilename: string);

		// IsEmpty:
		//        Indicates whether the document is Empty. It is automaticaly set after document Load, and
		//				cleared after document Clear. In addition to that, application can manualy alter this property
		//				when document is filled out in application, but there is no need to do so.
		// Loading:
		//				If true, means that loading of xml document is in progress
		// Saving:
		//				If true, means that saving of xml document is in progress
		property IsEmpty: boolean read FIsEmpty write FIsEmpty;
		property Loading: boolean read FLoading;
		property Saving: boolean read FSaving;
	published

		// PreserveWhiteSpace, MixedContent:
		//				See help for this properties on TXMLReader object. This values here serve as a starting
		//				options when new reader is created to load the document. Once the reader is created
		//				changing theese properties while loading is in progress will not affect reader, to
		//				affect the reader in read events change the reader's object properties directly.
		// FormatOutput
		//				See help for this property on TXMLWriter object. This value here serve as a starting
		//				option when new writer is created to save the document. Once the writer is created
		//				changing that property while saving is in progress will not affect writer, to
		//				affect the writer in write events change the writer's object property directly.
		// ProgressStep
		//				See help for this property on TXMLReader object.
		property PreserveWhiteSpace: boolean read FPreserveWhiteSpace write FPreserveWhiteSpace default false;
		property MixedContent: boolean read FMixedContent write FMixedContent default false;
		property ProgressStep: integer read FProgressStep write FProgressStep default 5;
		property FormatOutput: boolean read FFormatOutput write FFormatOutput default true;

		// OnReadDocumentStart:
		//				Called just before reader starts to load xml document
		// OnReadDocumentEnd:
		//				Called just after reader finished loading of xml document
		// OnReadNode:
		//        Called whenever node is completed while reading. It's called for every single node
		//				in the document. Reader.Node represents the node that is read. If application frees
		//				the specified node then it will not be included in document after loading.
		//				Note: Freeing element node in this event will cause subsequent OnReadElementEnd
		//				with invalid Reader.Node object so care has to be taken if objects are freed here
		//				and OnReadElementEnd is processed.
		// OnReadElementStart:
		//        Called whenever element node is created while reading and attributes are parsed and
		//				stored as child nodes. Element is still empty since content is still not parsed.
		//				Reader.Node represents the element node that is started. Application can not free
		//				specified node because that would generate exception but it can change/free it's attributes
		//				since they are allready parsed in.
		// OnReadElementEnd:
		//        Called whenever element node is completed while reading. It will be called after OnReadNode
		//				for this same Node.	Reader.Node represents the element node that is completed. If application frees
		//				the specified node then it will not be included in document after loading.
		// OnReadProgress:
		//        Called whenever progress changes. Progress can be accessed through Reader.ProgressPercent
		//				and represents percent value from 0..100 of reading progress.
		//
		// Note:	if exception is raised in any of OnReadXXX events, then parsing is stopped, Document property will
		//				contain xml objects that are read up to that point and exception will be in raised state
		//				after LoadFromXXX method. No try/except blocks are implemented in XMLReader.
		//				If you want to allow user interaction while document is reading then call Application.ProceessMessages
		//				somewhere in OnReadXXX events.
		property OnReadDocumentStart: TXMLReaderEvent read FOnReadDocumentStart write FOnReadDocumentStart;
		property OnReadDocumentEnd: TXMLReaderEvent read FOnReadDocumentEnd write FOnReadDocumentEnd;
		property OnReadNode: TXMLReaderEvent read FOnReadNode write FOnReadNode;
		property OnReadElementStart: TXMLReaderEvent read FOnReadElementStart write FOnReadElementStart;
		property OnReadElementEnd: TXMLReaderEvent read FOnReadElementEnd write FOnReadElementEnd;
		property OnReadProgress: TXMLReaderEvent read FOnReadProgress write FOnReadProgress;

		// OnWriteDocumentStart:
		//				Called just before writer starts to write xml document
		// OnWriteDocumentEnd:
		//				Called just after writer finished writing of xml document
		// OnWriteElementStart:
		//        Called whenever element node is about to be written. Writer.Node represents the element node
		//				that is about to be written.
		// OnWriteElementEnd:
		//        Called whenever element node is completed while writing. Writer.Node represents the element
		//				node that is completed.
		// OnWriteProgress:
		//        Called whenever progress changes. Progress can be accessed through Writer.ProgressPercent
		//				and represents percent value from 0..100 of writing progress.
		property OnWriteDocumentStart: TXMLWriterEvent read FOnWriteDocumentStart write FOnWriteDocumentStart;
		property OnWriteDocumentEnd: TXMLWriterEvent read FOnWriteDocumentEnd write FOnWriteDocumentEnd;
		property OnWriteElementStart: TXMLWriterEvent read FOnWriteElementStart write FOnWriteElementStart;
		property OnWriteElementEnd: TXMLWriterEvent read FOnWriteElementEnd write FOnWriteElementEnd;
		property OnWriteProgress: TXMLWriterEvent read FOnWriteProgress write FOnWriteProgress;
	end;

implementation

uses
	Math, StrLib; //, UTrace;

const
	LetterSet: CharSet = ['A'..'Z', 'a'..'z'];
	WhitespaceSet: CharSet = [' ', #9, #10, #13];
	CombiningCharSet: CharSet = [];
	DigitSet: CharSet = ['0'..'9'];
	ExtenderSet: CharSet = [#$b7];
	FirstNameCharSet: CharSet = ['_', ':', 'A'..'Z', 'a'..'z'];
var
	NameCharSet: CharSet;

{ TXBuf }

procedure TXBuf.Init;
begin
	FSize := 0;
	FPtr := nil;
end;

procedure TXBuf.Clear;
begin
	if FPtr <> nil then
		begin
			//DebugInfo('TXBuf.Clear: FreeMem - before');
			FreeMem(FPtr);
			//DebugInfo('TXBuf.Clear: FreeMem - after');
			FPtr := nil;
		end;
	FSize := 0;
end;

procedure TXBuf.CopyFrom(SourcePtr: PChar; SourceSize: integer);
begin
	SetSize(SourceSize);
	if SourceSize > 0 then
		Move(SourcePtr^, FPtr^, SourceSize);
end;

procedure TXBuf.AppendFrom(SourcePtr: PChar; SourceSize: integer);
var
	FOldSize: integer;
begin
	if SourceSize > 0 then
		begin
			FOldSize := FSize;
			SetSize(FSize+SourceSize);
			Move(SourcePtr^, (FPtr+FOldSize)^, SourceSize);
		end;
end;

function TXBuf.GetAsString: string;
begin
	if FSize <= 0 then
		result := ''
	else
		begin
			SetLength(result, FSize);
			Move(FPtr^, result[1], FSize);
		end;
end;

procedure TXBuf.SetAsString(value: string);
begin
	Clear;
	FSize := Length(value);
	if FSize > 0 then
		begin
			GetMem(FPtr, FSize);
			Move(value[1], FPtr^, FSize);
		end;
end;

procedure TXBuf.SetSize(value: integer);
//var
//	FNewPtr: PChar;
begin
	if FSize <> value then
		begin
			//DebugInfo('TXBuf.SetSize: ReallocMem - before');
			//DebugInfoFmt('  Size: %d, Ptr: %s', [FSize, string(FPtr)]);

			(*
			if Value = 0 then
				Clear
			else
				begin
					GetMem(FNewPtr, value);
					if FSize > 0 then
						Move(FPtr^, FNewPtr^, Math.Min(FSize, value));
					if FPtr <> nil then
						FreeMem(FPtr);
					FPtr := FNewPtr;
					FSize := value;
				end;
			*)
			FSize := value;
			ReallocMem(FPtr, FSize);

			//DebugInfo('TXBuf.SetSize: ReallocMem - after');
		end;
end;

(*
{$ifdef OBJBUF1}
function TrimChars(const AString: TXBuf; TrimChars: CharSet): TXBuf;
{$else}
function TrimChars(const AString: string; TrimChars: CharSet): string;
{$endif OBJBUF1}
var
	i, b, e: Integer;
	c: char;
	trim: boolean;
begin
{$ifdef OBJBUF1}
	e := AString.Size;
{$else}
	e := Length(AString);
{$endif OBJBUF1}
	b := 1;
	trim := false;
	for i := 1 to e do
		begin
{$ifdef OBJBUF1}
			c := AString.Ptr[i-1];
{$else}
			c := AString[i];
{$endif OBJBUF1}
			if c in TrimChars then
				begin
					b := i + 1;
					trim := true;
				end
			else
				Break;
		end;
	for i := e downto b do
		begin
{$ifdef OBJBUF1}
			c := AString.Ptr[i-1];
{$else}
			c := AString[i];
{$endif OBJBUF1}
			if c in TrimChars then
				begin
					e := i - 1;
					trim := true;
				end
			else
				Break;
		end;
{$ifdef OBJBUF1}
	if b > e then
		Result.Init
	else if trim then
		begin
			Result.FPtr := AString.FPtr + (b-1);
			Result.FSize := e - b + 1;
		end
	else
		Result := AString;
{$else}
	if b > e then
		Result := ''
	else if trim then
		Result := Copy(AString, b, e - b + 1)
	else
		Result := AString;
{$endif OBJBUF1}
end;
*)

function TrimWhiteSpace(const AString: string): string;
var
	i, b, e: Integer;
	trim: boolean;
begin
	e := Length(AString);
	b := 1;
	trim := false;
	for i := 1 to e do
		begin
			if AString[i] in WhiteSpaceSet then
				begin
					b := i + 1;
					trim := true;
				end
			else
				Break;
		end;
	for i := e downto b do
		begin
			if AString[i] in WhiteSpaceSet then
				begin
					e := i - 1;
					trim := true;
				end
			else
				Break;
		end;
	if b > e then
		Result := ''
	else if trim then
		Result := Copy(AString, b, e - b + 1)
	else
		Result := AString;
end;

{ TXMLNode }

constructor TXMLNode.Create(const NodeType: TXMLNodeType);
begin
	inherited Create;
{$ifdef LIST0}
	FNodes := TList.Create;
{$endif LIST0}
	FNodeType := NodeType;
end;

destructor TXMLNode.Destroy;
var
{$ifdef LIST2}
	n: TXMLNode;
{$else}
	i: Integer;
{$endif LIST2}
begin
	//DebugInfo('TXMLNode.Destroy ' + Description);

{$ifdef PARENT1}
	if FParentNode <> nil then
		FParentNode.RemoveChild(Self);
{$endif PARENT1}

{$ifdef LIST0}
	i := FNodes.Count;
	while i > 0 do
		begin
			Dec(i);
{$ifdef PARENT1}
			Nodes[i].FParentNode := nil;
{$endif PARENT1}
			Nodes[i].Free;
		end;
	FNodes.Free;
{$endif LIST0}

{$ifdef LIST1}
	if FNodes <> nil then
		begin
			i := FNodes.Count;
			while i > 0 do
				begin
					Dec(i);
{$ifdef PARENT1}
					Nodes[i].FParentNode := nil;
{$endif PARENT1}
					Nodes[i].Free;
				end;
			FNodes.Free;
		end;
{$endif LIST1}

{$ifdef LIST2}
	while FLastChild <> nil do
		begin
			n := FLastChild;
			FLastChild := n.PreviousSibling;
			// Release the node
{$ifdef PARENT1}
			n.FParentNode := nil;
{$endif PARENT1}
			n.FPreviousSibling := nil;
			n.FNextSibling := nil;
			n.Free;
		end;
{$endif LIST2}

{$ifdef OBJBUF1}
	FNodeName.Clear;
	FNodeValue.Clear;
{$endif OBJBUF1}

	inherited;
end;

{$ifdef PARENT1}
procedure TXMLNode.SetParent(AParent: TXMLNode);
begin
	if FParentNode <> AParent then
		begin
			if FParentNode <> nil then
				FParentNode.RemoveChild(Self);
			if AParent <> nil then
				AParent.AppendChild(Self);
		end;
end;

function TXMLNode.GetOwnerDocument: TXMLNode;
begin
	if FNodeType = xntDocument then
		result := Self
	else if FParentNode = nil then
		result := nil
	else
		result := FParentNode.OwnerDocument;
end;

function TXMLNode.GetPath: string;
var
	n: TXMLNode;
begin
	result := '';
	n := ParentNode;
	while n <> nil do
		begin
			result := n.NodeName + '\' + result;
			n := n.ParentNode;
		end;
end;
{$endif PARENT1}

{$ifdef OBJBUF1}
function TXMLNode.GetNodeName: string;
begin
	result := FNodeName.AsString;
end;

procedure TXMLNode.SetNodeName(value: string);
begin
	FNodeName.AsString := value;
end;

function TXMLNode.GetNodeValue: string;
begin
	result := FNodeValue.AsString;
end;

procedure TXMLNode.SetNodeValue(value: string);
begin
	FNodeValue.AsString := value;
end;
{$endif OBJBUF1}

function TXMLNode.Description: String;
begin
	Result := NodeName + ': ' + TXMLNodeTypeStrings[NodeType];
	if NodeValue <> '' then
		Result := Result + ' = ' + NodeValue;
end;

{$ifndef LIST2}
function TXMLNode.GetChildNodeCount: Integer;
begin
{$ifdef LIST1}
	result := 0;
	if FNodes <> nil then
{$endif LIST1}
		Result := FNodes.Count;
end;

function TXMLNode.GetNodes(Index: Integer): TXMLNode;
begin
{$ifdef LIST1}
	Result := nil;
	if FNodes <> nil then
{$endif LIST1}
		Result := TXMLNode(FNodes[Index]);
end;
{$endif LIST2}

function TXMLNode.GetTotalNodeCount: Integer;
var
{$ifdef LIST2}
	n: TXMLNode;
{$else}
	i: Integer;
{$endif LIST2}
begin
	Result := 1;
{$ifdef LIST2}
	n := FFirstChild;
	while n <> nil do
		begin
			Result := Result + n.TotalNodeCount;
			n := n.FNextSibling;
		end;
{$else}
	for i := 0 to Pred(ChildNodeCount) do
		Result := Result + Nodes[i].TotalNodeCount;
{$endif LIST2}
end;

function TXMLNode.CreateChild(const NodeType: TXMLNodeType): TXMLNode;
begin
	result := TXMLNode.Create(NodeType);
	AppendChild(result);
end;

{$ifdef LIST2}
procedure TXMLNode.AppendChild(Node: TXMLNode);
begin
{$ifdef PARENT1}
	// if node belongs to other parent node remove the reference from the parent
	if Node.ParentNode <> nil then
		Node.ParentNode.RemoveChild(Node);
	Node.FParentNode := Self;
{$endif PARENT1}

	if FLastChild <> nil then
		begin
			FLastChild.FNextSibling := Node;
			Node.FPreviousSibling := FLastChild;
		end
	else
		begin
			FFirstChild := Node;
		end;
	FLastChild := Node;
	Inc(FChildNodeCount);
end;

function TXMLNode.RemoveChild(Node: TXMLNode): boolean;
begin
{$ifdef PARENT1}
	result := Node.FParentNode = self;
{$else}
	result := true;
{$endif PARENT1}

	if result then
		begin
{$ifdef PARENT1}
			Node.FParentNode := nil;
{$endif PARENT1}
			if FFirstChild = Node then
				FFirstChild := Node.FNextSibling;
			if FLastChild = Node then
				FLastChild := Node.FPreviousSibling;

			if Node.FNextSibling <> nil then
				begin
					Node.FNextSibling.FPreviousSibling := Node.FPreviousSibling;
					Node.FNextSibling := nil;
				end;
			if Node.FPreviousSibling <> nil then
				begin
					Node.FPreviousSibling.FNextSibling := Node.FNextSibling;
					Node.FPreviousSibling := nil;
				end;

			Dec(FChildNodeCount);
		end;
end;

{$else}

function TXMLNode.AppendChild(Node: TXMLNode): Integer;
begin
{$ifdef LIST1}
	// create node list when needed
	if FNodes = nil then
		FNodes := TList.Create;
	Node.FParentNode := Self;
{$endif LIST1}

{$ifdef PARENT1}
	// if node belongs to other parent node remove the reference from the parent
	if Node.ParentNode <> nil then
		Node.ParentNode.RemoveChild(Node);
{$endif PARENT1}

	Result := FNodes.Add(Node);
end;

function TXMLNode.RemoveChild(Node: TXMLNode): integer;
begin
{$ifdef LIST1}
	result := -1;
	if FNodes <> nil then
		begin
{$endif LIST1}
			result := FNodes.Remove(Node);
{$ifdef PARENT1}
			if result >= 0 then
				Node.FParentNode := nil;
{$endif PARENT1}
{$ifdef LIST1}
		end;
{$endif LIST1}
end;
function TXMLNode.Delete(Index: Integer): TXMLNode;
begin
{$ifdef LIST1}
	result := nil;
	if FNodes <> nil then
		begin
{$endif LIST1}
			result := TXMLNode(FNodes[Index]);
			FNodes.Delete(Index);
{$ifdef PARENT1}
			result.FParentNode := nil;
{$endif PARENT1}
{$ifdef LIST1}
		end;
{$endif LIST1}
end;

function TXMLNode.IndexOf(Node: TXMLNode): integer;
begin
{$ifdef LIST1}
	result := -1;
	if FNodes <> nil then
{$endif LIST1}
		result := FNodes.IndexOf(Node);
end;
{$endif LIST2}

function TXMLNode.HasChildNodes: boolean;
begin
  result := ChildNodeCount > 0;
end;

function TXMLNode.CloneNode(deep: boolean): TXMLNode;
var
{$ifndef LIST2}
	i: integer;
{$endif LIST2}
	n: TXMLNode;
begin
	result := TXMLNode.Create(NodeType);
	try
		result.NodeName := NodeName;
		result.NodeValue := NodeValue;

		if deep then
      begin
			{$ifdef LIST2}
				n := FirstChild;
				while n <> nil do
					begin
			{$else}
				for i := 0 to ChildNodeCount-1 do
					begin
						n := Nodes[i];
			{$endif LIST2}
						result.AppendChild(n.CloneNode(true));
			{$ifdef LIST2}
						n := n.NextSibling;
			{$endif LIST2}
					end;
			end;

	except
		result.Free;
		raise;
	end;
end;

function TXMLNode.FindChild(const NodeName: String; const NodeType: TXMLNodeType = xntElement;
	const Index: Integer = 0; const CreateMissing: boolean = false): TXMLNode;
var
{$ifndef LIST2}
	i: Integer;
{$endif LIST2}
	ThisIndex: Integer;
	n: TXMLNode;
begin
{$ifdef LIST1}
	// create node list when needed
	if FNodes = nil then
		FNodes := TList.Create;
{$endif LIST1}

	ThisIndex := -1;
{$ifdef LIST2}
	n := FFirstChild;
	while n <> nil do
		begin
{$else}
	for i := 0 to Pred(ChildNodeCount) do
		begin
			n := Nodes[i];
{$endif LIST2}
			if (n.NodeType = NodeType) and (n.NodeName = NodeName) then
				begin
					Inc(ThisIndex);
					if ThisIndex = Index then
						begin
							Result := n;
							Exit;
						end;
				end;
{$ifdef LIST2}
			n := n.FNextSibling;
{$endif LIST2}
		end;

	Result := nil;

	if CreateMissing then
		begin
			while ThisIndex < Index do
				begin
					Result := CreateChild(NodeType);
					Result.NodeName := NodeName;
					Inc(ThisIndex);
				end;
		end;
end;

function TXMLNode.FindChildExc(const NodeName: String; const NodeType: TXMLNodeType = xntElement;
  const Index: Integer = 0): TXMLNode;
begin
  Result := FindChild(NodeName, NodeType, Index);
  if Result = nil then
    raise EXmlError.CreateFmt('Failed to find XML %s %s, index %d',
        [TXMLNodeTypeStrings[NodeType], NodeName, Index]);
end;

function TXMLNode.GetAttributes(Name: string): TXMLNode;
begin
	result := FindChild(Name, xntAttribute, 0, true);
end;

function TXMLNode.GetElements(Name: string): TXMLNode;
begin
	result := FindChild(Name, xntElement, 0, true);
end;

function TXMLNode.GetData: string;
var
{$ifndef LIST2}
	i: integer;
{$endif LIST2}
	n: TXMLNode;
begin
	// To get Data value we have to add NodeValue with NodeValues of child text and CData nodes
	result := NodeValue;
{$ifdef LIST2}
	n := FirstChild;
	while n <> nil do
		begin
{$else}
	for i := 0 to ChildNodeCount-1 do
		begin
			n := Nodes[i];
{$endif LIST2}
			if n.NodeType in [xntText, xntCData] then
				result := result + n.NodeValue;
{$ifdef LIST2}
			n := n.NextSibling;
{$endif LIST2}
		end;
end;

procedure TXMLNode.SetData(value: string);
var
{$ifndef LIST2}
	i: integer;
{$endif LIST2}
	n, x: TXMLNode;
begin
	// To set Data value we have to delete child text and CData nodes and
	// put the value into NodeValue
{$ifdef LIST2}
	n := FirstChild;
	while n <> nil do
		begin
			if n.NodeType in [xntText, xntCData] then
				begin
					x := n.NextSibling;
					n.Free;
					n := x;
				end
			else
				n := n.NextSibling;
		end;
{$else}
	i := 0;
	while i < ChildNodeCount do
		begin
			n := Nodes[i];
			if n.NodeType in [xntText, xntCData] then
				n.Free
			else
				Inc(i);
		end;
{$endif LIST2}
	NodeValue := value;
end;

function GetNodeText(Node: TXMLNode; Exact: boolean): string;
var
{$ifndef LIST2}
	i: integer;
{$endif LIST2}
	n: TXMLNode;
	last_nt: TXMLNodeType;
	t: string;
begin
	// To get Text value we have to go along the object node tree structure
	// starting from this node and add all NodeValues for nodes of
	// xntElement, xntText or	xntCData type.
	result := Node.NodeValue;
	last_nt := xntDocument;
{$ifdef LIST2}
	n := Node.FirstChild;
	while n <> nil do
		begin
{$else}
	for i := 0 to Node.ChildNodeCount-1 do
		begin
			n := Node.Nodes[i];
{$endif LIST2}
			if n.NodeType in [xntElement, xntText, xntCData] then
				begin
					t := GetNodeText(n, Exact);
					if not Exact and (n.NodeType = xntElement) and (last_nt = xntElement) and (result <> '') and (t <> '') then
						result := result + ' ' + t
					else
						result := result + t;
					last_nt := n.NodeType;
				end;
{$ifdef LIST2}
			n := n.NextSibling;
{$endif LIST2}
		end;
end;

function TXMLNode.GetText: string;
begin
	result := TrimWhiteSpace(GetNodeText(self, false));
end;

function TXMLNode.GetExactText: string;
begin
	result := TrimWhiteSpace(GetNodeText(self, true));
end;

function TXMLNode.GetXml: string;
var
	w: TXMLStringWriter;
begin
	result := '';
	w := TXMLStringWriter.Create(result);
	try
		w.StartDoc;
		w.WriteNode(Self);
		w.EndDoc;
	finally
		w.Free;
	end;
end;

{ TXMLReader }

{$ifdef PARSEBUF1}
(*
function TXMLReader.ReadBufOpen: integer;
begin
	result := FReadBufPos;
end;
*)

procedure TXMLReader.ReadBufAdd(c: char);
begin
	if FReadBufPos >= FReadBuf.Size then
		FReadBuf.Size := FReadBuf.Size + 256;

	FReadBuf.Ptr[FReadBufPos] := c;
	Inc(FReadBufPos);
end;

function TXMLReader.ReadBufClose(p: integer): TXBuf;
begin
	Result.Init;
	if p < FReadBufPos then
		begin
			Result.FSize := FReadBufPos - p;
			Result.FPtr := FReadBuf.Ptr + p;
			FReadBufPos := p;
		end;
end;
{$endif PARSEBUF1}

{$ifdef UNREADBUF1}
procedure TXMLReader.UnReadChar(c: char);
begin
	if FUnReadBufPos >= FUnReadBuf.Size then
		FUnReadBuf.Size := FUnReadBuf.Size + 256;

	FUnReadBuf.Ptr[FUnReadBufPos] := CurChar;
	Inc(FUnReadBufPos);
	
	CurChar := c;
end;

procedure TXMLReader.UnReadString(p: PChar; count: integer);
begin
	while count > 0 do
		begin
			Dec(count);
			UnReadChar(p[count]);
		end;
end;
{$endif UNREADBUF1}

procedure TXMLReader.ReadChar;
begin
{$ifdef UNREADBUF1}
	if FUnReadBufPos > 0 then
		begin
			// There are characters in the unread buffer, so we get it from there
			Dec(FUnReadBufPos);
			CurChar := FUnReadBuf.Ptr[FUnReadBufPos];
		end
	else
		begin
			CurChar := GetChar;
			// Progress indicator support only if we read from the XML source
			// because progress is based on that not the content of unread buffer
			// which may be filled out by the application in certain circumstances
			// when we realize the parsing that way.
			Dec(FProgressPos);
			if FProgressPos = 0 then
				begin
					Inc(FProgressPercent, FProgressStep);
					FProgressPos := FProgressCount;
					DoProgress;
				end;
		end;
{$else}
	// Progress indicator support. Since ReadChar is virtual method, actual reading will be performed
	// in the overrided method so here we just keep track of where we are in the progress of loading.
	Dec(FProgressPos);
	if FProgressPos = 0 then
		begin
			Inc(FProgressPercent, FProgressStep);
			FProgressPos := FProgressCount;
			DoProgress;
		end;
{$endif UNREADBUF1}
end;

procedure TXMLReader.ReadCharAndCheckForEnd;
begin
	if EndOfData then
		ReportError('Unexpected end of data');
	ReadChar;
end;

procedure TXMLReader.DoNode(Node: TXMLNode);
begin
	if Assigned(FOnNode) then
		begin
			FNode := Node;
			FOnNode(Self);
		end;
end;

procedure TXMLReader.DoElementStart(Node: TXMLNode);
begin
	if Assigned(FOnElementStart) then
		begin
			FNode := Node;
			FOnElementStart(Self);
		end;
end;

procedure TXMLReader.DoElementEnd(Node: TXMLNode);
begin
	if Assigned(FOnElementEnd) then
		begin
			FNode := Node;
			FOnElementEnd(Self);
		end;
end;

procedure TXMLReader.DoProgress;
begin
	if Assigned(FOnProgress) then	FOnProgress(Self);
end;

function TXMLReader.DoesStreamMatchString(Str: String): Boolean;
var
	i: Integer;
{$ifdef UNREADBUF1}
	c: char;
{$else}
	State: TXMLReaderState;
{$endif UNREADBUF1}
begin
	if Str = '' then begin Result := False; Exit end;
{$ifdef UNREADBUF0}
	SaveState(State);
{$endif UNREADBUF0}
	i := 1;
	repeat
{$ifdef UNREADBUF1}
		c := CurChar;
{$endif UNREADBUF1}
		Result := Str[i] = CurChar;
		Inc(i);
		ReadChar;
	until (not Result) or EndOfData or (i > Length(Str));

	if not Result then
		begin
{$ifdef UNREADBUF1}
			UnReadChar(c);
			UnReadString(@Str[1], i-2);
{$else}
			RestoreState(State);
{$endif UNREADBUF1}
		end;
end;

function TXMLReader.ParseAttribute(Node: TXMLNode): Boolean;
var
{$ifdef PARSEBUF1}
	Name: TXBuf;
{$else}
	Name: String;
{$endif PARSEBUF1}
	AttValue: String;
	ThisNode: TXMLNode;
begin
	result := ParseName(Name);
	if not Result then Exit;
	ThisNode := Node.CreateChild(xntAttribute);
{$ifdef PARSEBUF1}
	{$ifdef OBJBUF1}
	ThisNode.FNodeName.CopyFrom(Name.Ptr, Name.Size);
	{$else}
	ThisNode.NodeName := Name.AsString;
	{$endif OBJBUF1}
{$else}
	ThisNode.NodeName := Name;
{$endif PARSEBUF1}
	if not ParseEq then
		ReportError('Missing equals sign between attribute and attribute value');
	if not ParseAttValue(AttValue) then
		ReportError('Attribute ''' + ThisNode.NodeName + ''' is missing value. A string literal was expected, but no opening quote character was found');
	ThisNode.NodeValue := AttValue;
	DoNode(ThisNode);
end;

function TXMLReader.ParseAttValue(var AttValue: String): Boolean;
var
	QuoteChar: Char;
begin
	Result := False;
	if not (CurChar in ['"', '''']) then Exit;
	QuoteChar := CurChar;
	ReadCharAndCheckForEnd;

	while CurChar <> QuoteChar do
		begin
			if CurChar = '<' then
				ReportError('The character ''' + CurChar + ''' cannot be used in an attribute value')
			else if CurChar = '&' then
				CurChar := ParseReference;
			AttValue := AttValue + CurChar;
			ReadCharAndCheckForEnd;
		end;

	ReadCharAndCheckForEnd;
	Result := True;
end;

procedure TXMLReader.ParseComment;
var
	Comment: String;
	Terminated: Boolean;
begin
	repeat
		Comment := Comment + CurChar;
		Terminated := Right(Comment, 3) = '-->';
		if EndOfData then Break;
		ReadChar;
	until Terminated;
	if Right(Comment, 4) = '--->' then
		ReportError('Incorrect syntax was used in a comment. Comments cannot end with ''--->''');
end;

procedure TXMLReader.ParseContent(Node: TXMLNode; EndTagName: string);

var
{$ifdef PARSEBUF1}
	p: integer;
	Data: TXBuf;
{$else}
	Data: string;
{$endif PARSEBUF1}

const
	CDataEndMax = 3;
var
	CDataEndPos: byte;

	procedure ParseCDSect;
	var
{$ifdef PARSEBUF1}
		FReadBufFreeze: integer;
		CData: TXBuf;
{$else}
		CData: String;
{$endif PARSEBUF1}
	begin
{$ifdef PARSEBUF1}
		FReadBufFreeze := FReadBufPos;
{$endif PARSEBUF1}
		CDataEndPos := 0;
		repeat
{$ifdef PARSEBUF1}
			ReadBufAdd(CurChar);
{$else}
			CData := CData + CurChar;
{$endif PARSEBUF1}
			if CurChar = ']' then
				begin
					if CDataEndPos < (CDataEndMax-1) then
						Inc(CDataEndPos);
				end
			else if (CurChar = '>') and (CDataEndPos = CDataEndMax-1) then
				begin
					Inc(CDataEndPos);
				end
			else
				CDataEndPos := 0;
			ReadCharAndCheckForEnd;
		until CDataEndPos = CDataEndMax;

		// Remove CDATA ending sequence
{$ifdef PARSEBUF1}
		Dec(FReadBufPos, CDataEndMax);
{$else}
		CData := Left(CData, Length(CData) - 3);
{$endif PARSEBUF1}

		if FMixedContent then
			begin
{$ifdef PARSEBUF1}
				CData := ReadBufClose(FReadBufFreeze);
{$endif PARSEBUF1}
				with Node.CreateChild(xntCData) do
					begin
{$ifdef PARSEBUF1}
						{$ifdef OBJBUF1}
						FNodeValue.CopyFrom(CData.Ptr, CData.Size);
						{$else}
						NodeValue := CData.AsString;
						{$endif OBJBUF1}
{$else}
						NodeValue := CData;
{$endif PARSEBUF1}
					end;
			end
		else
			begin
{$ifdef PARSEBUF1}
				// no special handling, value stays in the Data buffer
{$else}
				Data := Data + CData;
{$endif PARSEBUF1}
			end;
	end;

	procedure ParseCharData;
	var
		bWhiteSpace: boolean;
{$ifdef PARSEBUF1}
		FReadBufFreeze: integer;
		CharData: TXBuf;
{$else}
		CharData: string;
{$endif PARSEBUF0}
	begin
{$ifdef PARSEBUF1}
		FReadBufFreeze := FReadBufPos;
{$endif PARSEBUF1}
		bWhiteSpace := not FPreserveWhiteSpace;
		CDataEndPos := 0;
		repeat
			if CurChar = '&' then
				begin
{$ifdef PARSEBUF1}
					ReadBufAdd(ParseReference);
{$else}
					CharData := CharData + ParseReference;
{$endif PARSEBUF1}
					bWhiteSpace := false;
					CDataEndPos := 0;
				end
			else
				begin
{$ifdef PARSEBUF1}
					ReadBufAdd(CurChar);
{$else}
					CharData := CharData + CurChar;
{$endif PARSEBUF1}
					if bWhiteSpace and not (CurChar in WhiteSpaceSet) then
						bWhiteSpace := false;
					if CurChar = ']' then
						begin
							if CDataEndPos < (CDataEndMax-1) then
								Inc(CDataEndPos);
						end
					else if (CurChar = '>') and (CDataEndPos = CDataEndMax-1) then
						ReportError('The literal string '']]>'' is not allowed in element content')
					else
						CDataEndPos := 0;
					ReadChar;
				end;
		until CurChar in ['<', #0];

		// If char data contains only white space and if we don't preserve it
		// then we drop it, otherwise we just append it to the content data buffer
		if bWhiteSpace then
			begin
{$ifdef PARSEBUF1}
				FReadBufPos := FReadBufFreeze;
{$else}
				// no special handling, CharData discarded
{$endif PARSEBUF1}
			end
		else if FMixedContent then
			begin
{$ifdef PARSEBUF1}
				CharData := ReadBufClose(FReadBufFreeze);
{$endif PARSEBUF1}
				with Node.CreateChild(xntText) do
					begin
{$ifdef PARSEBUF1}
						{$ifdef OBJBUF1}
						FNodeValue.CopyFrom(CharData.Ptr, CharData.Size);
						{$else}
						NodeValue := CharData.AsString;
						{$endif OBJBUF1}
{$else}
						NodeValue := CharData;
{$endif PARSEBUF1}
					end;
			end
		else
			begin
{$ifdef PARSEBUF1}
				// no special handling, value stays in the Data buffer
{$else}
				Data := Data + CharData;
{$endif PARSEBUF1}
			end;
	end;

	procedure ParseETag;
	var
{$ifdef PARSEBUF1}
		Name: TXBuf;
{$else}
		Name: String;
{$endif PARSEBUF1}
	begin
		if not ParseName(Name) then
			ReportError('The following tag was not closed: ' + EndTagName);
{$ifdef PARSEBUF1}
		if Name.AsString <> EndTagName then
			ReportError('End tag ''' + Name.AsString + ''' does not match the start tag ''' + EndTagName + '''');
{$else}
		if Name <> EndTagName then
			ReportError('End tag ''' + Name + ''' does not match the start tag ''' + EndTagName + '''');
{$endif PARSEBUF1}
		ParseWhitespace;
		if CurChar <> '>' then
			ReportError('End tag '''+EndTagName+''' was missing the character ''>''');
		ReadChar;
	end;

var
	FoundContent: Boolean;
begin
{$ifdef PARSEBUF1}
	p := FReadBufPos;
{$endif PARSEBUF1}

	FoundContent := False;
	while not FoundContent do
		begin
			if CurChar = '<' then
				begin
					ReadCharAndCheckForEnd;
					case CurChar of
						'/':
							begin
								ReadCharAndCheckForEnd;
								ParseETag;
								FoundContent := True;
							end;
						'?':
							begin
								ReadCharAndCheckForEnd;
								ParseProcessingInstruction(Node);
							end;
						'!':
							begin
								ReadCharAndCheckForEnd;
								if DoesStreamMatchString('--') then
									ParseComment
								else if DoesStreamMatchString('[CDATA[') then
									ParseCDSect
								else
									ReportError('Unrecognized character sequence after <!');
							end;
						else
							ParseElement(Node);
					end;
				end
			else if CurChar = #0 then
				ReportError('Unexpected end of character stream')
			else
				ParseCharData;
		end;
{$ifdef PARSEBUF1}
	Data := ReadBufClose(p);
	if Data.Size > 0 then
		begin
		{$ifdef OBJBUF1}
		Node.FNodeValue.CopyFrom(Data.Ptr, Data.Size);
		{$else}
		Node.NodeValue := Data.AsString;
		{$endif OBJBUF1}
		end;
{$else}
	if Data <> '' then
		Node.NodeValue := Data;
{$endif PARSEBUF1}
end;

constructor TXMLReader.Create;
begin
	FPreserveWhiteSpace := false;
	FMixedContent := false;
	FProgressStep := 5;
end;

destructor TXMLReader.Destroy;
begin
{$ifdef PARSEBUF1}
	FReadBuf.Clear;
{$endif PARSEBUF1}

{$ifdef UNREADBUF1}
	FUnReadBuf.Clear;
{$endif UNREADBUF1}
	inherited;
end;

procedure TXMLReader.ParseDocument(Document: TXMLNode);

	procedure ParseXMLDecl;
	var
		ThisNode: TXMLNode;
	begin
		if not DoesStreamMatchString('<?xml ') then Exit;
		ThisNode := Document.CreateChild(xntXMLDeclaration);
		ThisNode.NodeName := 'xml';
		repeat
			ParseWhitespace;
		until not ParseAttribute(ThisNode);
		ParseWhitespace;
		if not DoesStreamMatchString('?>') then
			ReportError('XML declaration not terminated with "?>"');
		DoNode(ThisNode);
	end;

	procedure ParseDocTypeDecl;
	var
		ThisNode: TXMLNode;
{$ifdef PARSEBUF1}
		Name, Data: TXBuf;
		p: integer;
{$else}
		Name, Data: string;
{$endif PARSEBUF1}
		OpenCount: integer;
	begin
		if not DoesStreamMatchString('<!DOCTYPE') then Exit;
		ParseWhitespace;
		if not ParseName(Name) then
			ReportError('Document type does not contain valid name');

		ThisNode := Document.CreateChild(xntDocumentType);
{$ifdef PARSEBUF1}
	{$ifdef OBJBUF1}
		ThisNode.FNodeName.CopyFrom(Name.Ptr, Name.Size);
	{$else}
		ThisNode.NodeName := Name.AsString;
	{$endif OBJBUF1}
{$else}
		ThisNode.NodeName := Name;
{$endif PARSEBUF1}
		ParseWhitespace;

		OpenCount := 1;
		// Document type data is stored as-is, no parsing, just collect the text until the ending >
{$ifdef PARSEBUF1}
		p := FReadBufPos;
{$else}
		Data := '';
{$endif PARSEBUF1}
		repeat
			if CurChar = '>' then
				Dec(OpenCount)
			else if CurChar = '<' then
				Inc(OpenCount);
			if OpenCount > 0 then
				begin
{$ifdef PARSEBUF1}
					ReadBufAdd(CurChar);
{$else}
					Data := Data + CurChar;
{$endif PARSEBUF1}
				end;
			ReadCharAndCheckForEnd;
		until OpenCount = 0;
{$ifdef PARSEBUF1}
		Data := ReadBufClose(p);
{$endif PARSEBUF1}

{$ifdef PARSEBUF1}
	{$ifdef OBJBUF1}
		ThisNode.FNodeValue.CopyFrom(Data.Ptr, Data.Size);
	{$else}
		ThisNode.NodeValue := Data.AsString;
	{$endif OBJBUF1}
{$else}
		ThisNode.NodeValue := Data;
{$endif PARSEBUF1}
		//ReportError('DocTypeDecl parsing not implemented');
		DoNode(ThisNode);
	end;

	function ParseMisc: Boolean;
	begin
		Result := true;
		if ParseWhitespace then Exit;
		if DoesStreamMatchString('<!--') then
			begin
				ParseComment;
				Exit;
			end;
		if DoesStreamMatchString('<?') then
			begin
				ParseProcessingInstruction(Document);
				Exit;
			end;
		Result := false;
	end;

	function ParseProlog: Boolean;
	begin
		ParseWhitespace;
		ParseXMLDecl;
		while ParseMisc do
			;
		ParseDocTypeDecl;
		while ParseMisc do
			;
		Result := True;
	end;

begin
	//DebugInfo('ParseDocument Begin');
	FProgressPercent := 0;
	FProgressCount := (Size div 100) * FProgressStep;
	FProgressPos := FProgressCount;
	DoProgress;

	ReadCharAndCheckForEnd;
	if not ParseProlog then
		ReportError('Missing prolog');

	if CurChar <> '<' then
		ReportError('Missing root element');
	ReadCharAndCheckForEnd;
	ParseElement(Document);

	while ParseMisc do
		;
	//DebugInfo('ParseDocument End');
	//DebugInfoFmt('  FSetNameCount = %d, FSetDataCount = %d', [FSetNameCount, FSetDataCount]);

	FProgressPercent := 100;
	DoProgress;
end;

procedure TXMLReader.ParseElement(Node: TXMLNode);
var
{$ifdef PARSEBUF1}
	Name: TXBuf;
{$else}
	Name: string;
{$endif PARSEBUF1}
	ThisNode: TXMLNode;
	EndTagName: string;
begin
	if not ParseName(Name) then
		ReportError('Missing element name');

	ThisNode := Node.CreateChild(xntElement);
{$ifdef PARSEBUF1}
{$ifdef OBJBUF1}
	ThisNode.FNodeName.CopyFrom(Name.Ptr, Name.Size);
{$else}
	ThisNode.NodeName := Name.AsString;
{$endif OBJBUF1}
{$else}
	ThisNode.NodeName := Name;
{$endif PARSEBUF1}
	EndTagName := ThisNode.NodeName;

	repeat
		ParseWhitespace;
	until not ParseAttribute(ThisNode);

	DoElementStart(ThisNode);

	if CurChar = '/' then begin
		// EmptyElemTag
		ReadCharAndCheckForEnd;
		if CurChar <> '>' then
			ReportError('Tag ''' + ThisNode.NodeName + ''' not terminated properly');
		ReadCharAndCheckForEnd;
	end else if CurChar = '>' then begin
		// STag
		ReadCharAndCheckForEnd;
		ParseContent(ThisNode, EndTagName);
	end else
		ReportError('Tag ''' + ThisNode.NodeName + ''' not terminated properly');

	DoNode(ThisNode);
	DoElementEnd(ThisNode);
end;

function TXMLReader.ParseEq: Boolean;
begin
	ParseWhitespace;
	Result := CurChar = '=';
	ReadCharAndCheckForEnd;
	ParseWhitespace;
end;

{$ifdef PARSEBUF1}
function TXMLReader.ParseName(var Name: TXBuf): boolean;
var
	p: integer;
{$else}
function TXMLReader.ParseName(var Name: string): boolean;
{$endif PARSEBUF1}
begin
	result := CurChar in FirstNameCharSet;
	if not Result then Exit;
{$ifdef PARSEBUF1}
	p := FReadBufPos;
{$else}
	Name := '';
{$endif PARSEBUF1}
	repeat
{$ifdef PARSEBUF1}
		ReadBufAdd(CurChar);
{$else}
		Name := Name + CurChar;
{$endif PARSEBUF1}
		ReadCharAndCheckForEnd;
	until not (CurChar in NameCharSet);
{$ifdef PARSEBUF1}
	Name := ReadBufClose(p);
{$endif PARSEBUF1}
end;

procedure TXMLReader.ParseProcessingInstruction(Node: TXMLNode);
var
{$ifdef PARSEBUF1}
	Name: TXBuf;
{$else}
	Name: string;
{$endif PARSEBUF1}
	Chars: string;
	ThisNode: TXMLNode;
	Terminated: Boolean;
begin
	if not ParseName(Name) then
		ReportError('Processing instruction does not contain valid name');
{$ifdef PARSEBUF1}
	if Name.AsString = 'xml' then
{$else}
	if Name = 'xml' then
{$endif PARSEBUF1}
		ReportError('Processing instruction cannot have the reserved name "xml"');
	ThisNode := Node.CreateChild(xntProcessingInstruction);
{$ifdef PARSEBUF1}
{$ifdef OBJBUF1}
	ThisNode.FNodeName.CopyFrom(Name.Ptr, Name.Size);
{$else}
	ThisNode.NodeName := Name.AsString;
{$endif OBJBUF1}
{$else}
	ThisNode.NodeName := Name;
{$endif PARSEBUF1}
	ParseWhitespace;
	repeat
		Chars := Chars + CurChar;
		Terminated := Right(Chars, 2) = '?>';
		if EndOfData then break;
		ReadChar;
	until Terminated;
	if not Terminated then
		ReportError('Unterminated processing instruction');
	ThisNode.NodeValue := Left(Chars, Length(Chars) - 2);
	
	DoNode(ThisNode);
end;

function TXMLReader.ParseReference: Char;
var
	RefStr: String;
begin
	Result := #0; // silence compiler warning
	if CurChar <> '&' then
		ReportError('All references must begin with "&"');
	ReadCharAndCheckForEnd;
	while CurChar <> ';' do begin
		RefStr := RefStr + CurChar;
		ReadCharAndCheckForEnd;
	end;
	if Left(RefStr, 2) = '#x' then
		try
			Result := Chr(StrToInt('$' + Right(RefStr, Length(RefStr) - 2)));
		except
			ReportError('Invalid integer in reference &' + RefStr + ';');
		end
	else if Left(RefStr, 1) = '#' then
		try
			Result := Chr(StrToInt(Right(RefStr, Length(RefStr) - 1)));
		except
			ReportError('Invalid integer in reference &' + RefStr + ';');
		end
	else if RefStr = 'amp' then
		Result := '&'
	else if RefStr = 'lt' then
		Result := '<'
	else if RefStr = 'gt' then
		Result := '>'
	else if RefStr = 'quot' then
		Result := '"'
	else if RefStr = 'apos' then
		Result := ''''
	else
		ReportError('Unrecognized reference');

	ReadCharAndCheckForEnd;
end;

function TXMLReader.ParseWhitespace: Boolean;
begin
	Result := False;
	while (CurChar in WhitespaceSet) do begin
		Result := True;
    ReadChar;
		if EndOfData then Exit;
  end;
end;

procedure TXMLReader.ReportError(Err: String);
begin
	raise EXmlError.Create(Err);
end;

{ TXMLStringReader }

constructor TXMLStringReader.Create(AData: string);
begin
	inherited Create;
	FData := AData;
	FPosition := 0;
end;

function TXMLStringReader.GetSize: integer;
begin
	result := length(FData);
end;

function TXMLStringReader.GetPosition: integer;
begin
	result := FPosition;
end;

function TXMLStringReader.EndOfData: Boolean;
begin
	Result := FPosition >= Length(FData);
end;

{$ifdef UNREADBUF1}
function TXMLStringReader.GetChar: char;
begin
	if FPosition >= Length(FData) then
		result := #0
	else
		begin
			Inc(FPosition);
			result := FData[FPosition];
		end
end;

{$else}

procedure TXMLStringReader.ReadChar;
begin
	inherited;
	if FPosition >= Length(FData) then
		CurChar := #0
	else
		begin
			Inc(FPosition);
			CurChar := FData[FPosition];
		end
end;

procedure TXMLStringReader.SaveState(var State: TXMLReaderState);
begin
	State.Position := FPosition;
	State.CurChar := CurChar;
end;

procedure TXMLStringReader.RestoreState(var State: TXMLReaderState);
begin
	FPosition := State.Position;
	CurChar := State.CurChar;
end;
{$endif UNREADBUF1}

{ TXMLStreamReader }

constructor TXMLStreamReader.Create(AStream: TStream);
begin
	inherited Create;
	FCacheLen := 0;
	FBehind := 0;
	FStream := AStream;
	FSize := AStream.Size;
	FPosition := AStream.Position;
end;

function TXMLStreamReader.GetSize: integer;
begin
	result := FSize;
end;

function TXMLStreamReader.GetPosition: integer;
begin
	result := FPosition - FBehind;
end;

function TXMLStreamReader.EndOfData: Boolean;
begin
	Result := (FPosition - FBehind) >= FSize;
end;

{$ifdef UNREADBUF1}
function TXMLStreamReader.GetChar: char;
begin
	if FBehind > 0 then
		begin
			result := FCacheBuf[FCacheLen-FBehind];
			Dec(FBehind);
		end
	else
		begin
			if FPosition >= FSize then
				result := #0
			else
				begin
					// Load cache as much as posible
					FCacheLen := FStream.Read(FCacheBuf[0], XMLStreamReaderCacheSize);
					FBehind := FCacheLen;
					FPosition := FStream.Position;

					result := FCacheBuf[FCacheLen-FBehind];
					Dec(FBehind);
				end;
		end;
end;

{$else}

procedure TXMLStreamReader.ReadChar;
var
	Saved: integer;
begin
	inherited;
	if FBehind > 0 then
		begin
			CurChar := FCacheBuf[FCacheLen-FBehind];
			Dec(FBehind);
		end
	else
		begin
			if FPosition >= FSize then
				CurChar := #0
			else
				begin
					// Load as much as posible, but before that move remaining buffer
					// to the front so that we keep at most XMLStreamReaderCacheSave of last chars in the buffer

					Saved := FCacheLen;
					if Saved > XMLStreamReaderCacheSave then
						Saved := XMLStreamReaderCacheSave;
					if Saved > 0 then
						System.Move(FCacheBuf[FCacheLen-Saved], FCacheBuf[0], Saved);

					FBehind := FStream.Read(FCacheBuf[Saved], (XMLStreamReaderCacheSize-Saved) and XMLStreamReaderLoadMask);
					//DebugInfo('TXMLStreamReader.ReadChar: FStream.Read');
					//DebugInfo('  FStream.Position = ' + IntToStr(FPosition));
					//DebugInfo('  nRead = ' + IntToStr(FBehind));
					FCacheLen := Saved + FBehind;
					FPosition := FStream.Position;

					CurChar := FCacheBuf[FCacheLen-FBehind];
					Dec(FBehind);
				end;
		end;
end;

procedure TXMLStreamReader.SaveState(var State: TXMLReaderState);
begin
	State.Position := FPosition - FBehind;
	State.CurChar := CurChar;
end;

procedure TXMLStreamReader.RestoreState(var State: TXMLReaderState);
begin
	FBehind := FPosition - State.Position;
	if (FBehind < 0) or (FBehind > FCacheLen) then
		begin
			FBehind := 0;
			FCacheLen := 0;
			FStream.Position := State.Position;
			FPosition := State.Position;
			//DebugInfo('TXMLStreamReader.RestoreState: position behind cash!!!');
		end;

	CurChar := State.CurChar;
end;
{$endif UNREADBUF1}

{ TXMLFileReader }

constructor TXMLFileReader.Create(AFileName: string);
begin
	inherited Create(TFileStream.Create(AFileName, fmOpenRead));
	FFileName := AFileName;
end;

destructor TXMLFileReader.Destroy;
begin
	if Stream <> nil then
		Stream.Free;
	inherited;
end;

{ TXMLWriter }

constructor TXMLWriter.Create;
begin
	inherited;
	FFormatOutput := true;
	//FProgressStep := 5;
end;

destructor TXMLWriter.Destroy;
begin
	FWriteBuf.Clear;
	inherited;
end;

procedure TXMLWriter.StartDoc;
begin
	FWriteBuf.Size := XMLWriterCacheSize;
	FWriteBufPos := 0;
	FIndentation := 0;
	FFirstLine := true;

	FPosition := 0;
	FTotal := 0;
	FCount := 0;
	FProgressPercent := 0;
end;

procedure TXMLWriter.EndDoc;
begin
	if FWriteBufPos > 0 then
		WriteBuffer(FWriteBuf.Ptr, FWriteBufPos);
	FWriteBuf.Clear;
end;

procedure TXMLWriter.WriteBuffer(Buffer: PChar; Count: integer);
begin
	// Progress count support
	Inc(FPosition, Count);
	if FTotal <> 0 then
		FProgressPercent := (100 * FCount) div FTotal;
	DoProgress;
end;

procedure TXMLWriter.WriteChar(c: char);
begin
	if FWriteBufPos >= XMLWriterCacheSize then
		begin
			WriteBuffer(FWriteBuf.Ptr, FWriteBufPos);
			FWriteBufPos := 0;
		end;
	FWriteBuf.Ptr[FWriteBufPos] := c;
	Inc(FWriteBufPos);
end;

procedure TXMLWriter.DoElementStart(Node: TXMLNode);
begin
	if Assigned(FOnElementStart) then
		begin
			FNode := Node;
			FOnElementStart(Self);
		end;
end;

procedure TXMLWriter.DoElementEnd(Node: TXMLNode);
begin
	if Assigned(FOnElementEnd) then
		begin
			FNode := Node;
			FOnElementEnd(Self);
		end;
end;

procedure TXMLWriter.DoProgress;
begin
	if Assigned(FOnProgress) then	FOnProgress(Self);
end;

procedure TXMLWriter.WriteStr(const Value: String);
var
	done: boolean;
	Size: integer;
begin
  if Value = '' then Exit;
  
	done := false;
	Size := Length(Value);
	while not done do
		if FWriteBufPos + Size <= XMLWriterCacheSize then
			begin
				Move(Value[1], FWriteBuf.Ptr[FWriteBufPos], Size);
				Inc(FWriteBufPos, Size);
				done := true;
			end
		else if FWriteBufPos > 0 then
			begin
				WriteBuffer(FWriteBuf.Ptr, FWriteBufPos);
				FWriteBufPos := 0;
			end
		else
			begin
				WriteBuffer(@Value[1], Size);
				done := true;
			end;
end;

procedure TXMLWriter.Write(const Value: String);
var
	Size: integer;
	i: Integer;
	c: char;
begin
	Size := Length(Value);
	i := 1;
	while i <= Size do
		begin
			c := Value[i];
			case c of
				'<': 	WriteStr('&lt;');
				'>': 	WriteStr('&gt;');
				'''':	WriteStr('&apos;');
				'"': 	WriteStr('&quot;');
				'&':	WriteStr('&amp;');
			else
				(*if Ord(c) > 127 then
					WriteStr('&#' + IntToStr(Ord(Value[i])) + ';')
				else*)
					WriteChar(c);
			end;
			Inc(i);
		end;
end;

procedure TXMLWriter.WriteLineEnd;
var
	i: integer;
begin
	if FFormatOutput then
		begin
			if FFirstLine then
				FFirstLine := false
			else
				begin
					WriteChar(#13);
					WriteChar(#10);
					for i := 1 to FIndentation do
						begin
							WriteChar(' ');
							WriteChar(' ');
						end;
				end
		end;
end;

procedure TXMLWriter.WriteStandaloneDocumentDeclaration;
begin
	WriteStr('<?xml version="1.0" standalone="yes"?>');
end;

procedure TXMLWriter.WriteTag(const Name: String; TagType: TXMLTagType;
	Open: Boolean);
begin
	if TagType in [xttStarting, xttStartingAndEnding] then
		begin
			WriteLineEnd;
			WriteChar('<');
		end
	else if TagType in [xttEnding, xttEndingSameLine] then
		begin
			Dec(FIndentation);
			if TagType = xttEnding then
				WriteLineEnd;
			WriteStr('</');
		end;
	WriteStr(Name);
	if not Open then
		WriteTagClose(TagType);
end;

procedure TXMLWriter.WriteTagClose(TagType: TXMLTagType);
begin
	if TagType = xttStartingAndEnding then
		WriteStr('/>')
	else
		begin
			WriteStr('>');
			if TagType = xttStarting then
				Inc(FIndentation);
		end;
end;

procedure TXMLWriter.WriteCDATA(Data: string);
const
	CDataEndMax = 3;
var
	CDataEndPos: byte;
	i, size: integer;
	c: char;
begin
	size := Length(Data);
	if size > 0 then
		begin
			CDataEndPos := 1;
			WriteStr('<![CDATA[');
			for i := 1 to Size do
				begin
					c := Data[i];
					if c = ']' then
						begin
							if CDataEndPos < (CDataEndMax-1) then
								Inc(CDataEndPos);
							WriteChar(c);
						end
					else if (c = '>') and (CDataEndPos = CDataEndMax-1) then
						begin
							CDataEndPos := 0;
							WriteStr(']]>');
							Write(c);
							WriteStr('<![CDATA[');
						end
					else
						begin
							CDataEndPos := 0;
							WriteChar(c);
						end;
				end;
			WriteStr(']]>');
		end;
end;

procedure TXMLWriter.WriteData(const Name, Data: String);
begin
	WriteTag(Name);
	Write(Data);
	WriteTag(Name, xttEndingSameLine);
end;

procedure TXMLWriter.WriteTagParam(const ParamName, Value: String);
begin
	WriteChar(' ');
	WriteStr(ParamName);
	WriteChar('=');
	WriteChar('"');
	Write(Value);
	WriteChar('"');
end;

procedure TXMLWriter.WriteNode(Node: TXMLNode);
var
	is_empty: boolean;
	has_elem: boolean;
{$ifndef LIST2}
	i: integer;
{$endif LIST2}
	n: TXMLNode;
begin
	// If WriteNode called immediately after StartDoc
	// then we have to set total nodes count so that
	// progress can be maintained
	if FTotal = 0 then
		FTotal := Node.TotalNodeCount;

	if Node.NodeType = xntXMLDeclaration then
		begin
			WriteLineEnd;
			WriteStr('<?xml');
{$ifdef LIST2}
			n := Node.FirstChild;
			while n <> nil do
				begin
{$else}
			for i := 0 to Node.ChildNodeCount-1 do
				begin
					n := Node.Nodes[i];
{$endif LIST2}
					if n.NodeType = xntAttribute then
						WriteTagParam(n.NodeName, n.NodeValue);
{$ifdef LIST2}
					n := n.NextSibling;
{$endif LIST2}
					Inc(FCount);
				end;
			WriteStr('?>');
		end
	else if Node.NodeType = xntProcessingInstruction then
		begin
			WriteLineEnd;
			WriteStr('<?' + Node.NodeName + ' ' + Node.NodeValue + '?>');
		end
	else if Node.NodeType in [xntDocument, xntElement] then
		begin
			if Node.NodeType = xntElement then
				DoElementStart(Node);

			is_empty := Node.NodeValue = '';
			has_elem := Node.NodeType = xntDocument;

			if Node.NodeType = xntElement then
				begin
					WriteTag(Node.NodeName, xttStarting, true);

{$ifdef LIST2}
					n := Node.FirstChild;
					while n <> nil do
						begin
{$else}
					for i := 0 to Node.ChildNodeCount-1 do
						begin
							n := Node.Nodes[i];
{$endif LIST2}
							if (n.NodeType = xntAttribute) then
								begin
									WriteTagParam(n.NodeName, n.NodeValue);
									Inc(FCount);
								end
							else if n.NodeType in [xntText, xntCData] then
								is_empty := false
							else
								has_elem := true;
{$ifdef LIST2}
							n := n.NextSibling;
{$endif LIST2}
						end;

					if is_empty and not has_elem then
						WriteTagClose(xttStartingAndEnding)
					else
						begin
							WriteTagClose(xttStarting);
							if not is_empty then
								Write(Node.NodeValue);
						end;
				end;

{$ifdef LIST2}
			n := Node.FirstChild;
			while n <> nil do
				begin
{$else}
			for i := 0 to Node.ChildNodeCount-1 do
				begin
					n := Node.Nodes[i];
{$endif LIST2}
					if n.NodeType <> xntAttribute then
						begin
							if n.NodeType = xntText then
								begin
									Write(n.NodeValue);
									Inc(FCount);
								end
							else if n.NodeType = xntCData then
								begin
									WriteCDATA(n.NodeValue);
									Inc(FCount);
								end
							else
								WriteNode(n);
						end;
{$ifdef LIST2}
					n := n.NextSibling;
{$endif LIST2}
				end;

			if (Node.NodeType = xntElement) and (not is_empty or has_elem) then
				begin
					if has_elem then
						WriteTag(Node.NodeName, xttEnding)
					else
						WriteTag(Node.NodeName, xttEndingSameLine);
				end;

			if Node.NodeType = xntElement then
				DoElementEnd(Node);
		end;

	Inc(FCount);
end;

{ TXMLStringWriter }

constructor TXMLStringWriter.Create(var AString: string);
begin
	inherited Create;
	FString := @AString;
end;

procedure TXMLStringWriter.WriteBuffer(Buffer: PChar; Count: integer);
var
	Size: integer;
begin
	inherited;
	Size := Length(FString^);
	SetLength(FString^, Size+Count);
	Move(Buffer[0], FString^[Size+1], Count);
end;

{ TXMLStreamWriter }

constructor TXMLStreamWriter.Create(AStream: TStream);
begin
	inherited Create;
	FStream := AStream;
end;

procedure TXMLStreamWriter.WriteBuffer(Buffer: PChar; Count: integer);
begin
	inherited;
	FStream.Write(Buffer[0], Count);
end;

{ TXMLFileWriter }

constructor TXMLFileWriter.Create(AFileName: string);
begin
	inherited Create(TFileStream.Create(AFileName, fmCreate or fmShareDenyWrite));
	FFileName := AFileName;
end;

destructor TXMLFileWriter.Destroy;
begin
	if Stream <> nil then
		Stream.Free;
	inherited;
end;

{ TXMLDocument }

constructor TXMLDocument.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
	FDocument := TXMLNode.Create(xntDocument);
	FIsEmpty := true;
	FFormatOutput := true;
	FProgressStep := 5;
end;

destructor TXMLDocument.Destroy;
begin
	if FDocument <> nil then
		FDocument.Free;
	inherited;
end;

procedure TXMLDocument.Clear;
begin
	CheckBusy;
	if FDocument <> nil then
		FDocument.Free;
	FDocument := TXMLNode.Create(xntDocument);
	FIsEmpty := true;
end;

(*
function TXMLDocument.createNode(NodeType: TXMLNodeType; const Name: String): TXMLNode;
begin
	result := TXMLNode.Create(NodeType);
	result.NodeName := Name;
end;

function TXMLDocument.createElement(const Name: String): TXMLNode;
begin
	result := createNode(xntElement, Name);
end;

function TXMLDocument.createAttribute(const Name: String): TXMLNode;
begin
	result := createNode(xntAttribute, Name);
end;
*)

procedure	TXMLDocument.DoReadDocumentStart(Reader: TXMLReader);
begin
	if Assigned(FOnReadDocumentStart) then
		FOnReadDocumentStart(Self, Reader);
end;

procedure	TXMLDocument.DoReadDocumentEnd(Reader: TXMLReader);
begin
	if Assigned(FOnReadDocumentEnd) then
		FOnReadDocumentEnd(Self, Reader);
end;

procedure	TXMLDocument.DoReadNode(Sender: TObject);
begin
	FOnReadNode(Self, TXMLReader(Sender));
end;

procedure	TXMLDocument.DoReadElementStart(Sender: TObject);
begin
	FOnReadElementStart(Self, TXMLReader(Sender));
end;

procedure	TXMLDocument.DoReadElementEnd(Sender: TObject);
begin
	FOnReadElementEnd(Self, TXMLReader(Sender));
end;

procedure	TXMLDocument.DoReadProgress(Sender: TObject);
begin
	FOnReadProgress(Self, TXMLReader(Sender));
end;

procedure	TXMLDocument.DoWriteDocumentStart(Writer: TXMLWriter);
begin
	if Assigned(FOnWriteDocumentStart) then
		FOnWriteDocumentStart(Self, Writer);
end;

procedure	TXMLDocument.DoWriteDocumentEnd(Writer: TXMLWriter);
begin
	if Assigned(FOnWriteDocumentEnd) then
		FOnWriteDocumentEnd(Self, Writer);
end;

procedure	TXMLDocument.DoWriteElementStart(Sender: TObject);
begin
	FOnWriteElementStart(Self, TXMLWriter(Sender));
end;

procedure	TXMLDocument.DoWriteElementEnd(Sender: TObject);
begin
	FOnWriteElementEnd(Self, TXMLWriter(Sender));
end;

procedure	TXMLDocument.DoWriteProgress(Sender: TObject);
begin
	FOnWriteProgress(Self, TXMLWriter(Sender));
end;

procedure TXMLDocument.CheckBusy;
begin
	if Loading then
		raise EXmlError.Create(Name + ':TXMLDocument is still busy loading document');
	if Saving then
		raise EXmlError.Create(Name + ':TXMLDocument is still busy saving document');
end;

procedure TXMLDocument.LoadFromReader(r: TXMLReader);
begin
	// Clear will call CheckBusy...
	Clear;
	FLoading := true;
	try
		r.PreserveWhiteSpace := FPreserveWhiteSpace;
		r.MixedContent := FMixedContent;
		r.ProgressStep := FProgressStep;

		if Assigned(FOnReadNode) then r.OnNode := DoReadNode;
		if Assigned(FOnReadElementStart) then r.OnElementStart := DoReadElementStart;
		if Assigned(FOnReadElementEnd) then r.OnElementEnd := DoReadElementEnd;
		if Assigned(FOnReadProgress) then r.OnProgress := DoReadProgress;

		DoReadDocumentStart(r);
		r.ParseDocument(FDocument);
		DoReadDocumentEnd(r);

		FIsEmpty := false;
	finally
		FLoading := false;
		r.Free;
	end;
end;

procedure TXMLDocument.LoadFromString(xmlstring: string);
begin
	LoadFromReader(TXMLStringReader.Create(xmlstring));
end;

procedure TXMLDocument.LoadFromStream(xmlstream: TStream);
begin
	LoadFromReader(TXMLStreamReader.Create(xmlstream));
end;

procedure TXMLDocument.LoadFromFile(xmlfilename: string);
begin
	LoadFromReader(TXMLFileReader.Create(xmlfilename));
end;

procedure TXMLDocument.SaveToWriter(w: TXMLWriter);
begin
	CheckBusy;
	FSaving := true;
	try
		w.FormatOutput := FFormatOutput;
		//w.ProgressStep := FProgressStep;

		if Assigned(FOnWriteElementStart) then w.OnElementStart := DoWriteElementStart;
		if Assigned(FOnWriteElementEnd) then w.OnElementEnd := DoWriteElementEnd;
		if Assigned(FOnWriteProgress) then w.OnProgress := DoWriteProgress;

		w.StartDoc;
		DoWriteDocumentStart(w);
		w.WriteNode(FDocument);
		w.EndDoc;
		DoWriteDocumentEnd(w);
	finally
		FSaving := false;
		w.Free;
	end;
end;

function TXMLDocument.SaveToString: string;
begin
	result := '';
	SaveToWriter(TXMLStringWriter.Create(result));
end;

procedure TXMLDocument.SaveToStream(xmlstream: TStream);
begin
	SaveToWriter(TXMLStreamWriter.Create(xmlstream));
end;

procedure TXMLDocument.SaveToFile(xmlfilename: string);
begin
	SaveToWriter(TXMLFileWriter.Create(xmlfilename));
end;

initialization
	NameCharSet := LetterSet + CombiningCharSet + DigitSet + ExtenderSet
			+ ['.', '-', '_', ':'];
end.

