unit Main;

interface

  procedure Dump(const PEFile: string; const LogFile: string);
  function InsertTabs(const Line: string; TabSize: Integer = 8):string;

implementation
uses
  Windows,
  ImageHlp,
  ExeHdr,
  Classes,
  FileSys,
  SysUtils,
  FileUtils,
  ConsoleApp;

function InsertTabs(const Line: string; TabSize: Integer = 8):string;

  function CountOfChar(const Line: string; StartIndex: Integer; Value: Char): Integer;
  begin
    Result := 0;
    while (StartIndex <= Length(Line)) and (Line[StartIndex] = Value) do begin
      Inc(Result);
      Inc(StartIndex)
    end
  end;

  function ReplaceSpaces(Column: Integer; CountOfSpaces: Integer): string;
  var
    TabColumn: Integer;
    SpaceCount: Integer;
  begin
    case CountOfSpaces of
    0: Result := '';
    1: Result := ' ';
    else
      Result := '';
      while CountOfSpaces > 0 do begin
        TabColumn := Column + 1;
        while (TabColumn - 1) mod TabSize <> 0 do
          Inc(TabColumn);
        SpaceCount := TabColumn - Column;
        if SpaceCount <= CountOfSpaces then begin
          Result := Result + #9;
          Dec(CountOfSpaces, SpaceCount);
          Column := TabColumn
        end else begin
          Result := Result + StringOfChar(' ', CountOfSpaces);
          exit
        end
      end;
    end;
  end;

  function ExpandTabs(const line: string): string;
  var
    i: integer;
  begin
    Result := line;
    while true do begin
      i := Pos(#9, Result);
      if i = 0 then break;
      Delete(Result, i, 1);
      Insert(StringOfChar(' ', 1 + TabSize - (i mod TabSize)), Result, i)
    end
  end;

var
  I: Integer;
  CountOfSpaces: Integer;
  Buffer: string;
begin
  Result := '';
  Buffer := ExpandTabs(Line);

  I := 1;
  while I <= Length(Buffer) do
    if Buffer[I] = ' ' then begin
      CountOfSpaces := CountOfChar(Buffer, I, ' ');
      Result := Result + ReplaceSpaces(I, CountOfSpaces);
      Inc(I, CountOfSpaces)
    end else begin
      Result := Result + Buffer[I];
      Inc(I)
    end
end;

type
  TNameOverrides = class(TStringList)
  public
    Tag: Integer;
  end;

  TPipeReader = class
  private
    InDisassembly: Boolean;
    InHeader: Boolean;
    SkipEmptyLines: Boolean;
    ProbablyInDumpHeader: Boolean;
    IgnorePaddingBytes: Boolean;
  private
    ProgressCounter: Cardinal;
    LastProgress: Cardinal;
    procedure Progress(Increment: Cardinal);
  private
    NameOverrides: TNameOverrides;
    Stream: TStream;
    LineBuffer: string;
    LineBufferQueue: TStrings;
  public
    constructor Create(const AFileName: string; ANameOverrides: TNameOverrides);
    destructor Destroy; override;
    procedure OnNewLine(Process: THandle; Buffer:PChar; Count:integer);
  end;

const
  MinOverrideCodeLength = 4;  { TUNE: }

function AddName(NameOverrides: TNameOverrides; const Name: string; BufferSize: Integer): string;
var
  I: Integer;
  Code: Integer;
begin
  { only limited number of codes can be encoded using the MinOverrideCodeLength.
    these short codes will be used only for very short buffers }
  if NameOverrides.Count = 0 then begin
    { allocated buffers for short codes }
    for I := 0 to (1 shl (4 * (MinOverrideCodeLength - 2))) - 1 do
      NameOverrides.Add('');
    NameOverrides.Tag := 0    { set 0 as first available short code }
  end;
  if BufferSize = MinOverrideCodeLength then begin
    { use short code }
    Code := NameOverrides.Tag;
    NameOverrides[Code] := Name;
    NameOverrides.Tag := NameOverrides.Tag + 1;
    // TODO: check for tag overflow
  end else begin
    Code := NameOverrides.Add(Name)
  end;

  Result := IntToHex(Code, 8);
  while (Length(Result) > 1) and (Result[1] = '0') do
    Delete(Result, 1, 1);

  Result := '{' + UpperCase(Result) + '}';
end;

procedure TPipeReader.Progress(Increment: Cardinal);
var
  NewProgress: Cardinal;
begin
  Inc(ProgressCounter, Increment);
  NewProgress := ProgressCounter div (2 { TUNE: step size in KB } * 1024);
  if NewProgress <> LastProgress then begin
    Write(NewProgress, #8#8#8#8#8#8);
    LastProgress := NewProgress
  end
end;

constructor TPipeReader.Create(const AFileName: string; ANameOverrides: TNameOverrides);
begin
  inherited Create;
  NameOverrides := ANameOverrides;
  Stream := TFileStream.Create(AFileName, fmCreate);
  Stream.Free;
  Stream := TFileStream.Create(AFileName, fmOpenReadWrite or fmShareDenyNone);
  Stream := TCachedStream.Create(Stream, True, 10 { TUNE: } * 4096, fmOpenWrite);
  InHeader := True;
  InDisassembly := False;
  SkipEmptyLines := False;
  ProbablyInDumpHeader := False;
  IgnorePaddingBytes := False;
  LineBufferQueue := TStringList.Create;
  LineBuffer := ''
end;

destructor TPipeReader.Destroy;
var
  I: Integer;
  Line: string;
begin
  { TODO: flush last buffer reusing the filter logic }
  begin
    if Assigned(LineBufferQueue) then
      if not ProbablyInDumpHeader then
        if LineBufferQueue.Count > 0 then begin
          for I := 0 to LineBufferQueue.Count - 1 do
            Line := LineBufferQueue[I];
            Stream.WriteBuffer(Line[1], Length(Line))
        end;

    if LineBuffer <> '' then begin
      Stream.WriteBuffer(LineBuffer[1], Length(LineBuffer));
    end;
  end;

  FreeAndNil(Stream);
  inherited
end;

procedure TPipeReader.OnNewLine(Process: THandle; Buffer:PChar; Count:integer);
var
  I: Integer;
label
  tryNext;

  procedure SendLine(Line: string; const Tail: string);

    { determines if line contains memory dump instruction }
    function IsDumpLine(const Line: string): Boolean;
    begin
      Result := False;
      if Line = '' then exit;

      if Copy(Line, 32, 9) = ' dd      ' then begin
        Result := True;
        exit
      end;

      if Length(Line) < 61 then exit;

      if Line[10] = ' ' then exit;

      if Line[1] = ' ' then exit;
      if Line[12] <> ' ' then exit;
      Result := True
    end;

    { determines if line contains just label (function label, offset) }
    function IsLabelLine(const Line: string): Boolean;
    begin
      Result := False;
      if Line = '' then exit;
      if Pos('                    ', Line) = 9 then
        if Line[29] <> ' ' then begin
          Result := True;
          exit
        end
    end;

    { determines if line contains just comment }
    function IsCommentLine(const Line: string): Boolean;
    begin
      Result := False;
      if Line = '' then exit;
      if Pos('                    ', Line) = 9 then
        if Line[29] = ' ' then
          if Line[57] = ';' then begin
            Result := True;
            exit
          end
    end;

    { determines if line contains control flow terminator (jump, return) }
    function IsFlowTerminatorLine(const Line: string): Boolean;
    begin
      Result := False;
      if Line = '' then exit;
      { is a bytecode present? }
      if Line[10] = ' ' then exit;
      if Trim(Copy(Line, 33, 4)) = 'ret' then begin
        Result := True;
        exit
      end;
      if Trim(Copy(Line, 33, 4)) = 'jmp' then begin
        Result := True;
        exit
      end;
    end;

    { determines if line contains typical opcode used to after control
      flow terminated to pad next label to dword or quadword boundary }
    function IsDummyFillerLine(const Line: string): Boolean;
    begin
      Result := False;
      if Pos('90                     nop', Line) <> 0 then begin
        Result := True;
        exit
      end;
      if Pos('CC                     int     3', Line) <> 0 then begin
        Result := True;
        exit
      end;
    end;

    procedure DoSendLine(const Line: string);
    var
      Buffer: string;
    begin
      if Line <> '' then begin
        Buffer := InsertTabs(Line, 8);
        Stream.WriteBuffer(Buffer[1], Length(Buffer));
      end
    end;

    procedure SendLine;
    begin
      DoSendLine(Line + Tail);
    end;

    procedure SendBufferQueue;
    var
      I: Integer;
    begin
      for I := 0 to LineBufferQueue.Count - 1 do
        DoSendLine(LineBufferQueue[I]);
      LineBufferQueue.Clear
    end;

    function ReplaceNameOverrides(const Line: string): string;
    var
      I: Integer;
      J: Integer;
      Code: string;
    label
      retry, seekNext;
    begin
      Result := Line;
      { replace NameOverrides codes with actual overriden names }
      I := Pos('{', Result);
    retry:
      if I <> 0 then begin
        Code := '';
        J := I + 1;
        while J <= Length(Result) do begin
          if Result[J] = '{' then begin
            { false alarm, lets restart }
            I := J;
            goto retry
          end;
          if Result[J] = '}' then begin
            if Code = '' then goto seekNext;
            Result := Copy(Result, 1, I - 1) + NameOverrides[StrToInt('$' + Code)] + Copy(Result, J + 1, MaxInt);
            J := I;
            goto seekNext
          end;
          if Result[J] in ['0'..'9', 'A'..'F'] then begin
            Code := Code + Result[J];
            Inc(J)
          end else begin
            { previous bracked was false alarm. lets look for next starting bracket }
            seekNext:
            I := J + 1;
            while I <= Length(Result) do begin
              if Result[I] = '{' then goto retry;
              Inc(I)
            end;
            break;
          end
        end;
      end;
    end;

  label
    doDisplay;
  begin
    Progress(Length(Line) + Length(Tail));

    if InHeader then begin
      if Pos('IAT Entry', Line) <> 0 then begin
        InHeader := False;
        exit
      end;
      goto doDisplay
    end else
    if InDisassembly then begin
      if SkipEmptyLines then begin
        if Trim(Line) = '' then exit;
        SkipEmptyLines := False
      end;

      if ProbablyInDumpHeader then begin
        { find out if the dump header continues, can be discarded or must be flushed }
        if IsCommentLine(Line) then begin
          { header probably continues }
          LineBufferQueue.Add(Line + Tail);
          exit
        end;
        if IsDumpLine(Line) then begin
          { yes, it was dump header. discard it }
          SkipEmptyLines := True;
          LineBufferQueue.Clear;
          ProbablyInDumpHeader := False;
          exit
        end;
        { it was not dump header, flush it }
        SendBufferQueue;
        ProbablyInDumpHeader := False
      end else begin
        if IsLabelLine(Line) then begin
          { probably dump header starts }
          LineBufferQueue.Add(ReplaceNameOverrides(Line) + Tail);
          ProbablyInDumpHeader := True;
          exit
        end;
      end;

      if IsFlowTerminatorLine(Line) then begin
        IgnorePaddingBytes := True;
      end else if IsDummyFillerLine(Line) then begin
        if IgnorePaddingBytes then
          exit
      end else
        IgnorePaddingBytes := False;

      if IsDumpLine(Line) then begin
        SkipEmptyLines := True;
        exit;
      end;

    doDisplay:
      { replace NameOverrides codes with actual overriden names }
      Line := ReplaceNameOverrides(Line);

      if Trim(Line) = '' then
        SkipEmptyLines := True;

      SendLine
    end else begin
      if Pos('Disassembly', Line) <> 0 then begin
        InDisassembly := True;
        SkipEmptyLines := False;

        SendLine
      end
    end
  end;

begin
  if Count > 0 then begin
    I := Length(LineBuffer);
    SetLength(LineBuffer, I + Count);
    Move(Buffer^, LineBuffer[I + 1], Count);
  tryNext:
    I := Pos(#13#10, LineBuffer);
    if I <> 0 then begin
      SendLine(Copy(LineBuffer, 1, I - 1), #13#10);
      Delete(LineBuffer, 1, (I -1) + 2);
      goto tryNext
    end;
  end
end;

procedure DumpPE(const DumpPEFileName: string; const PEFile: string; const LogFile: string; NameOverrides: TNameOverrides);
var
  Reader: TPipeReader;
begin
  Reader := TPipeReader.Create(LogFile, NameOverrides);
  try
    ExecConsoleApp(DumpPEFileName, '-disasm "' + PEFile + '"', '', nil, Reader.OnNewLine);
  finally
    Reader.Free
  end;
end;

function GetLibraryFileName(const PEFile: string): string;
var
  Lib: THandle;
  FileName: array [0..MAX_PATH] of Char;
begin
  Lib := LoadLibraryEx(PChar(PEFile), 0, DONT_RESOLVE_DLL_REFERENCES);
  if Lib = 0 then
    RaiseLastWin32Error;
  try
    if GetModuleFileName(Lib, @FileName, SizeOf(FileName)) = 0 then
      RaiseLastWin32Error;
    Result := StrPas(@FileName)
  finally
    FreeLibrary(Lib)
  end
end;

function UnDecoratedName(const SymbolName: string): string;
var
  Buffer: array [0..1024 { TUNE: }] of Char;

  function ReduceNestedTemplates(const SymbolName: string): string;
  var
    I: Integer;
    J: Integer;
    Level: Integer;
  begin
    Result := SymbolName;
    I := Pos('<', Result);
    if I <> 0 then begin
      if Copy(Result, I - (Length('operator<') - 1), Length('operator<')) = 'operator<' then begin
        { this is not template start bracket. let's skip it }
        J := I + 1; { skip 1 more letter for the case operator<< }
        Result := Copy(Result, 1, J) + ReduceNestedTemplates(Copy(Result, J + 1, MaxInt));
        exit
      end;

      Level := 1;
      J := I + 1;
      while True do begin
        if J > Length(Result) then begin
          { incomplete translation }
          exit;
        end;

        case Result[J] of
        '<':
          begin
            Inc(Level);
          end;

        '>':
          begin
            Dec(Level);
            if Level = 0 then begin
              Result := Copy(Result, 1, I - 1) + '<...>' + ReduceNestedTemplates(Copy(Result, J + 1, MaxInt));
              break;
            end;
          end;

        end;
        Inc(J)
      end
    end;
  end;

var
  I: Integer;  
begin
  Result := SymbolName;
  if SymbolName = '' then exit;
  if SymbolName[1] = '?' then begin
    { This is C++ style decorated name. See http://www.kegel.com/mangle.html for more }
    if UnDecorateSymbolName(
        PChar(SymbolName),
        @Buffer[0],
        SizeOf(Buffer),
        UNDNAME_COMPLETE or
        UNDNAME_NO_ACCESS_SPECIFIERS or
        UNDNAME_NO_MEMBER_TYPE or
        UNDNAME_NO_SPECIAL_SYMS or
        UNDNAME_NO_THROW_SIGNATURES or
        UNDNAME_NO_THISTYPE or
        UNDNAME_NO_MS_THISTYPE or
        UNDNAME_NO_LEADING_UNDERSCORES or
        UNDNAME_NO_ALLOCATION_MODEL or
        UNDNAME_32_BIT_DECODE) <> 0 then
      Result := StrPas(Buffer);

      { Remove some uninteresting parts }
      Result := StringReplace(Result, 'thiscall ', ' ', [rfReplaceAll]);

      { Simplify nested templates }
      Result := ReduceNestedTemplates(Result);

      Result := StringReplace(Result, ' *', '*', [rfReplaceAll]);
      Result := StringReplace(Result, ' &', '&', [rfReplaceAll]);

      Result := Trim(Result);
      while True do begin
        I := Pos('  ', Result);
        if I = 0 then break;
        Delete(Result, I, 1)
      end;
  end;
end;

procedure FixLongNames(const OldPEFile: string; const FixedPEFile: string; NameOverrides: TNameOverrides);
var
  ImageBase: Cardinal;
  img: PImageNTHeaders;
  Stream: TStream;

  function RVAToFileOffset(RVA: Cardinal): Cardinal;

    function FindSectionAt(RVA: Cardinal): PImageSectionHeader;
    var
      i: integer;
    begin
      Result := PImageSectionHeader(@PChar(img)[sizeof(img^.Signature) + sizeof(img^.FileHeader) + img^.FileHeader.SizeOfOptionalHeader]);
      for i := 1 to img^.FileHeader.NumberOfSections do
        if (RVA >= Cardinal(Result^.VirtualAddress)) and (RVA < Cardinal(Result^.VirtualAddress) + Cardinal(Result^.Misc.VirtualSize)) then
          exit
        else
          Result := PImageSectionHeader(@PChar(Result)[sizeof(TImageSectionHeader)]);
      Result := nil
    end;

  var
    sect: PImageSectionHeader;
  begin
    sect := FindSectionAt(RVA);
    Assert(sect <> nil);
    Result := Cardinal(sect^.PointerToRawData) + RVA - Cardinal(sect^.VirtualAddress)
  end;

  procedure ReadRaw(RVA: Cardinal; var Buffer; Size: Cardinal);
  begin
    Stream.Position := RVAToFileOffset(RVA);
    Stream.ReadBuffer(Buffer, Size);
  end;

  procedure WriteRaw(RVA: Cardinal; var Buffer; Size: Cardinal);
  begin
    Stream.Position := RVAToFileOffset(RVA);
    Stream.WriteBuffer(Buffer, Size);
  end;

  procedure FixImports(BaseRVA: Cardinal; Size: Cardinal);
  type
    PThunkTable = ^TThunkTable;
    TThunkTable = array[0..10000] of pointer;
  var
    imp: PImageImportDescriptor;
    impSize: Cardinal;
    firstThunkPtr: Cardinal;
    nameThunks: PThunkTable;
    thunkPtr: Cardinal;
    nameBuf: packed record
      Hint: Word;
      Name: packed array[0..800 { TUNE: }] of Char;
    end;
    name: PImageImportByName;
    n: Integer;
    OldImportName: string;
    NewImportName: string;
    LibraryName: string;
  begin
    if BaseRVA = 0 then exit;
    imp := PImageImportDescriptor(ImageBase + BaseRVA);
    impSize := Size;
    while (impSize > 0) and (imp^.Name <> 0) do begin
      LibraryName := ChangeFileExt(StrPas(PChar(Cardinal(imp^.Name) + ImageBase)), '');
      firstThunkPtr := Cardinal(imp^.OriginalFirstThunk);
      if firstThunkPtr = 0 then
        firstThunkPtr := Cardinal(imp^.FirstThunk);
      nameThunks := PThunkTable(firstThunkPtr + ImageBase);

      n := 0;
      while nameThunks^[n] <> nil do begin
        { nameThunks array may be already converted to direct function pointers.
          we need original name tables so we'll read the nameThunk table from
          disk }
        ReadRaw(Cardinal(@nameThunks^[n]) - ImageBase, thunkPtr, SizeOf(thunkPtr));

        if (thunkPtr and image_Ordinal_Flag) = 0 then
        begin
          ReadRaw(thunkPtr, nameBuf, SizeOf(nameBuf));
          name := PImageImportByName(@nameBuf);
          OldImportName := StrPas(name^.Name);

          if Length(OldImportName) >= MinOverrideCodeLength then begin
            { we'll better do the renaming }
            NewImportName := AddName(NameOverrides, LibraryName + '.' + UnDecoratedName(OldImportName), Length(OldImportName)) + #0;

            Move(NewImportName[1], name^.Name, Length(NewImportName));
            WriteRaw(thunkPtr, nameBuf, SizeOf(name^.Hint) + Length(NewImportName));
          end;

        end;

        Inc(n)
      end;

      Dec(impSize, SizeOf(TImageImportDescriptor));
      imp := PImageImportDescriptor(@PChar(imp)[SizeOf(TImageImportDescriptor)]);
    end
  end;

  procedure FixExports(BaseRVA: Cardinal);
  type
    PNameTable = ^TNameTable;
    TNameTable = array[0..10000] of pointer;
  var
    exp: PImageExportDirectory;
    n: Integer;
    namePtr: Cardinal;
    nameBuf: packed record
      Name: packed array[0..800 { TUNE: }] of Char;
    end;
    OldExportName: string;
    NewExportName: string;
  begin
    if BaseRVA = 0 then exit;
    exp := PImageExportDirectory(ImageBase + BaseRVA);
    for n := 0 to exp^.NumberOfNames - 1 do begin
      namePtr := Cardinal(PNameTable(Cardinal(exp^.AddressOfNames) + ImageBase)[n]);
      ReadRaw(namePtr, nameBuf, SizeOf(nameBuf));
      OldExportName := StrPas(nameBuf.Name);

      if Length(OldExportName) >= MinOverrideCodeLength then begin
        { we'll better do the renaming }
        NewExportName := UnDecoratedName(OldExportName);
        if NewExportName <> OldExportName then begin
          NewExportName := AddName(NameOverrides, NewExportName, Length(OldExportName)) + #0;

          Move(NewExportName[1], nameBuf.Name, Length(NewExportName));
          WriteRaw(namePtr, nameBuf, 0 + Length(NewExportName));
        end;
      end;
    end;
  end;

begin
  ImageBase := LoadLibraryEx(PChar(OldPEFile), 0, DONT_RESOLVE_DLL_REFERENCES);
  if ImageBase = 0 then
    RaiseLastWin32Error;
  try
    img := PImageNTHeaders(Cardinal(PImageDosHeader(ImageBase)^.e_lfanew) + ImageBase);

    OSFileSystem.FileSetAttr(FixedPEFile, 0);   // clear write protection attributes
    Stream := TFileStream.Create(FixedPEFile, fmOpenReadWrite);
    try
      with img^.OptionalHeader.DataDirectory[image_Directory_Entry_Import] do
        FixImports(VirtualAddress, Size);
      FixExports(img^.OptionalHeader.DataDirectory[image_Directory_Entry_Export].VirtualAddress);
    finally
      Stream.Free
    end
  finally
    FreeLibrary(ImageBase)
  end
end;

procedure ExtractFile(const ID: string; const Path: string);
var
  Src, Dst: TStream;
begin
  Src := TResourceStream.Create(hInstance, ID, RT_RCDATA);
  try
    Dst := TFileStream.Create(Path, fmCreate);
    try
      Dst.CopyFrom(Src, 0)
    finally
      Dst.Free
    end
  finally
    Src.Free
  end
end;

{$R DumpPEDat.res}

procedure Dump(const PEFile: string; const LogFile: string);
var
  OriginalFileName: string;
  WorkingFileName: string;
  TemporaryFolder: string;
  NameOverrides: TNameOverrides;
  DumpPEFileName: string;
begin
  OriginalFileName := GetLibraryFileName(PEFile);
  TemporaryFolder := OSFileSystem.GetTempFileName('\');
  try
    WorkingFileName := FileUtils.CreateFileName(OSFileSystem, TemporaryFolder, ExtractFileName(OriginalFileName), False);
    if not Windows.CopyFile(PChar(OriginalFileName), PChar(WorkingFileName), False) then
      RaiseLastWin32Error;

    NameOverrides := TNameOverrides.Create;
    try
      FixLongNames(OriginalFileName, WorkingFileName, NameOverrides);

      DumpPEFileName := FileUtils.CreateFileName(OSFileSystem, TemporaryFolder, 'DumpPE.exe', False);
      ExtractFile('DUMPPE', DumpPEFileName);

      DumpPE(DumpPEFileName, WorkingFileName, LogFile, NameOverrides)
    finally
      NameOverrides.Free
    end
  finally
    FileUtils.EraseFolder(TemporaryFolder)
  end;
end;

end.
